/*
 *  Bincache (C) 2007-2008
 *  Licensed under the GNU GPL v2, or later versions of the GPL at your option.
 */
#define BINCACHE_BUFFER_LEN 4096
#ifndef WIN32
# define _XOPEN_SOURCE 500
# define _BSD_SOURCE
# include <sys/time.h>
# include <time.h>
# include <sys/socket.h>
# include <sys/select.h>
# include <netinet/in.h>
# include <netdb.h>
#else
# include <winsock2.h>
# include <ws2tcpip.h>
# include <limits.h>
#endif
#include <gnutls/gnutls.h>
#include <gnutls/x509.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#ifndef WIN32
#include <sys/ioctl.h>
#include <termios.h>
#endif
#include <signal.h>
#include <dirent.h>

#ifdef WIN32
#define STDC_HEADERS
#include "regex.h"
#define sockerrno WSAGetLastError()
#define makedir mkdir
#define SHUT_RDWR SD_BOTH
#define EINPROGRESS WSAEINPROGRESS
#else
#define __USE_MISC
#include <arpa/inet.h>
#include <regex.h>
#define O_BINARY 0
#define sockerrno errno
#define makedir(x) mkdir(x,0777)
#endif
#include "bincache.h"

#ifdef WIN32
const char *sockstrerror(int err)
{
  static char buf[128];
  switch(err) {
    case WSAECONNREFUSED:
      return "ECONNREFUSED";
    default:
      snprintf(&buf[0],127,"Unknown error %d",err);
      buf[127]='\0';
      return &buf[0];
  }
}
#else
#define sockstrerror strerror
#endif

struct brokentime {
        int year, month, day, hour, minute;
        time_t stamp;
};

struct addrport {
        unsigned char a,b,c,d;
        unsigned int port;
};

struct buflet {
        int len;
        int pos;
        char *s;
};

static struct range_interest {
        char *group;
	unsigned long minl;
	unsigned long maxl;
        //struct brokentime min;
        //struct brokentime max;
        time_t last_activity; /* Clock time of last activity in this range.  TODO: use for expiring stale ranges. */
        } *g_ranges_desired;
static int g_num_ranges_desired;
static struct range_interest *g_posessed_ranges;
static int g_num_posessed_ranges;
static int g_num_slave_connections;

typedef enum {
        PROTOCOL_DEFAULT,
        PROTOCOL_AUTO_NNTP,
        // TODO: Auto-detection of best NNTP protocol version.
        PROTOCOL_NNTP,
        PROTOCOL_NNTP_SSL,
        PROTOCOL_CACHEPLEX_TRACKER,
        PROTOCOL_BINCACHE_PEER,
} protocol_enum_type;

static struct peer_connection {
  int refcount;
  protocol_enum_type prot;
  char *name;
  char *username;
  char *password;
  char *message_id; /* Message ID of current message being requested/handled with that peer. */
  struct sockaddr_in conn;
  enum {
        PEER_UNCONNECTED,
        PEER_CLOSED,
        PEER_TCP_CONNECTING,
        PEER_TLS_HANDSHAKING,
        PEER_WRITING_BANNER,
        PEER_READING_BANNER,
        PEER_READY_FOR_COMMAND,
        PEER_WAITING_FOR_NEWSGROUPS_HDR,
        PEER_WAITING_FOR_DATE_HDR,
        PEER_WAITING_FOR_STAT_RESPONSE,
        PEER_WAITING_FOR_MODE_READER_RESP,
        PEER_WAITING_FOR_AUTHINFO_USER_RESP,
        PEER_WAITING_FOR_AUTHINFO_PASS_RESP,
        PEER_WAITING_FOR_HEAD_RESP,
        PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK,
	PEER_RECEIVING_DATE_HDR_BLOCK,
        PEER_RECEIVING_HEAD_BLOCK,
        PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT,
        PEER_DISCARDING_BLOCK,
        PEER_WAITING_FOR_BODY_RESP,
        PEER_RECEIVING_BODY_BLOCK,
        PEER_BIDI_READY_FOR_COMMAND,
	PEER_SENDING_ARTICLE_BODY,
	PEER_SENDING_ARTICLE_HEAD,
  } state;
  struct timeval lastcmdtime;
  int lag;
  int fd;
  char buf[BINCACHE_BUFFER_LEN];
  int bufstart, bufpos;
  int buflinelen;
  int isfullfeed;
  int tried_mode_reader;
  int tried_to_authenticate; /* -1: auth rejected; 1: auth accepted. */
  /* Some servers react to unsupported commands by closing the connection.  To
   * protect against this, a closed connection after a HDR newsgroups or HDR
   * date command is treated as an error - otherwise, a bincache client would
   * be locked out of such servers entirely.  However, this may lead to
   * spurious fallbacks to HEAD in high-error conditions where the connection
   * is closed for no good reason. */
  int hdr_not_supported;
  int hdr_newsgroups_not_supported;
  int hdr_date_not_supported;
  int headers_only; /* Determined at ini time, so can be stored in the peer
        struct rather than the master struct. */
  /* Marks whether or not the current block's buffer starts after a CRLF. */
  int block_at_crlf;
  // TODO bandwidth cap
  int trace; /* Also determined at ini time. */
  time_t lastread; // TODO For idle detection in future versions
  /* This stuff is all per-peer.  It would probably be rude to establish
   * multiple connections to a single bincache peer, and at any rate most
   * bincache peers are liable to be bandwidth-bound rather than 
   * lookup-bound.*/
  int num_ranges_desired_by_peer;
  struct range_interest *ranges_desired_by_peer;
  int num_ranges_to_ask_for;
  struct range_interest *ranges_to_ask_for;
  int num_ranges_asked_for;
  struct range_interest *ranges_asked_for;
  int num_lowpriqueue;
  struct buflet *lowpriqueue;
  int tls;
        /* 
         * Default is false.  True for bincache and cacheplex connections.
         * Will be true for NNTPS when supported. 
         */
  gnutls_certificate_credentials_t tls_credentials;
  gnutls_anon_client_credentials_t tls_anoncred;
  char *tls_server_cert;
  gnutls_session_t tls_session;

  int write_block_atlf;
  struct desired_article *receiving_article;
  struct desired_article *sending_article;
  int sending_article_rpos;
    /* TODO there are some places where an article lookup by message-ID
     * occurs.  Articles should be attached instead.
     */
  struct peer_connection *master; /* Null if no master, pointer to peer master
        otherwise. */
  int numslaves;
  int num;
} **g_peer_connections;
static int peer_serial_number;
static int g_num_peer_connections;
static struct peer_connection **g_cacheplex_connections;
static int g_num_cacheplex_connections;

//static gnutls_certificate_credentials_t g_x509_credentials;

struct client_connection;

static struct desired_article {
  int refcount;
  char *message_id;
  /* The following servers have been asked if they have this message. */
  struct peer_connection **asked;
  int numasked;
  /* The following servers have this message. */
  struct peer_connection **have;
  int numhave;
  /* The following servers want this message. */
  struct peer_connection **want;
  int numwant;
  struct peer_connection **donthave;
  int numdonthave;
  // XXX cannot deal with multiple downloads of the same article, must fix. 
  int numleechers;
  struct client_connection **leechers;
  int unavailable;
  char *newsgroups_header;
  int newsgroups_header_len;
  int dateparsed;
  int year, month, day, hour, minute;
  time_t timestamp;
  char *head;
  int headlen;
  int headerscomplete;
  int articlefd; /* File descriptor message body is being written to. */
  int articlefd_wpos; /* Position in FD to write to. */
  int articlefd_bodystart;
  int bodycomplete; /* Entire body has been downloaded. */
  int gettingbody;
  int articlelen;
  time_t last_activity;
} **g_desired_articles;

static int g_num_desired_articles;

static struct client_connection {
  int refcount;
  int fd;
  struct sockaddr peeraddr;
  socklen_t peeraddrlen;
  char buf[BINCACHE_BUFFER_LEN];
  int bufstart; /* Start of unparsed data. */
  int bufpos;
  enum {
        CLIENT_WAITING_FOR_COMMAND,
        CLIENT_CLOSED,
        CLIENT_WAITING_FOR_ARTICLE_AVAILABILITY, 
                /* Waiting for either one of the peers to announce it has
                 * the desired article or all to announce that they do not. */
        CLIENT_WAITING_FOR_ARTICLE_HEAD,
        CLIENT_WAITING_FOR_ARTICLE_BODY,
        CLIENT_SENDING_ARTICLE_HEAD,
        CLIENT_SENDING_ARTICLE_BODY,
        CLIENT_SENDING_ARTICLE,
        } state;
  struct desired_article *article;
//  int article_fd;
  char *writebuf;
  int writebufpos;
  int writebuflen;
  int articlefd_rpos;
  int write_block_atlf;
} **g_client_connections;
static int g_num_client_connections;

static struct desired_article **g_posessed_articles;
static int g_num_posessed_articles;

static struct peer_connection *g_default_server;
static fd_set g_watched_rfds;
static fd_set g_watched_wfds;
static int g_high_fd;
int g_nntp_listen_socket;
struct sockaddr g_nntp_listen_addr;
char *g_cache_dir = "messages"; // TODO set in ini file.
regex_t g_date_wdy_DMYHMS_Z;
regex_t g_date_DMYHMS_z; // Date: 30 Aug 2007 07:41:21 GMT

static int g_bincache_listen_port, g_bincache_real_listen_port;
struct sockaddr g_bincache_listen_addr;
int g_bincache_listen_socket = -1;
static gnutls_dh_params_t g_dh_params;
static gnutls_anon_server_credentials_t g_anon_credentials;

void peer_deref(struct peer_connection *peer);
void client_deref(struct client_connection *conn);

void breakable(void)
{
  /* This is just a marker to set breakpoints at. */
}

void article_free(struct desired_article *article)
{
  if (article->message_id) free(article->message_id);
  if (article->articlefd>=0) close(article->articlefd);
  if (article->asked) {
    while (article->numasked) {
      --article->numasked;
      peer_deref(article->asked[article->numasked]);
    }
    free(article->asked);
  }
  if (article->have) {
    while (article->numhave) {
      --article->numhave;
      peer_deref(article->have[article->numhave]);
    }
    free(article->have);
  }
  if (article->want) {
    while (article->numwant) {
      peer_deref(article->want[--article->numwant]);
    }
    free(article->want);
  }
  if (article->donthave) {
    while (article->numdonthave) {
      peer_deref(article->donthave[--article->numdonthave]);
    }
    free(article->donthave);
  }
  if (article->leechers) {
    while (article->numleechers) {
      client_deref(article->leechers[--article->numleechers]);
    }
    free(article->leechers);
  }
  if (article->newsgroups_header) free(article->newsgroups_header);
  if (article->head) free(article->head);
  free(article);
}

void forget_article(struct desired_article *article)
{
  assert(article->last_activity+300<time(NULL));
  assert(article->refcount==0);
  int x = g_num_desired_articles;
  while (x) {
    x--;
    if (g_desired_articles[x] == article) {
      article_free(article);
      g_desired_articles[x] = g_desired_articles[--g_num_desired_articles];
      return;
    }
  }
  x = g_num_posessed_articles;
  while (x) {
    x--;
    if (g_posessed_articles[x] == article) {
      article_free(article);
      g_posessed_articles[x] = g_posessed_articles[--g_num_posessed_articles];
      return;
    }
  }
}

void article_deref(struct desired_article *article)
{
  article->refcount--;
  if (article->refcount<=0) {
    if (article->refcount<0) fprintf(stderr,"bincache: article reference count underflow.\n");
    if (article->last_activity+300<time(NULL)) {
      forget_article(article);
    }
  }
}

void client_deref(struct client_connection *conn)
{
  conn->refcount--;
  if (conn->refcount<=0) {
    if (conn->refcount<0) {
      fprintf(stderr,"bincache: BUG: reference count underflow.\n");
    }
    if (conn->article) article_deref(conn->article);
    if (conn->writebuf) free(conn->writebuf);
    free(conn);
  } 
}

struct client_connection *client_ref(struct client_connection *conn)
{
  conn->refcount++;
  return conn;
}

struct desired_article *article_ref(struct desired_article *article)
{
  article->refcount++;
  return article;
}

void peer_close(struct peer_connection *conn)
{
  if (conn->fd >= 0) {
    close(conn->fd);
    FD_CLR(conn->fd, &g_watched_rfds);
    FD_CLR(conn->fd, &g_watched_wfds);
    conn->fd = -1;
  }
  conn->state = PEER_CLOSED;
}

int writes(int fd, const char *s)
{
  int slen = strlen(s);
  return send(fd,s,slen,0)==slen;
}

void peer_conn_writes(struct peer_connection *conn, const char *s)
{
  if (conn->trace) printf("bincache: %d:%s [C] %s",conn->num,conn->name, s);
  if (!conn->tls) {
    if (writes(conn->fd,s)) return;
  } else {
    int slen = strlen(s);
    if (gnutls_record_send(conn->tls_session,s,slen)==slen) return;
  }
  peer_close(conn);
}

int peer_conn_writebuf(struct peer_connection *conn, const char *s, int slen)
{
  if (!conn->tls) {
    return send(conn->fd,s,slen,0);
  } else {
    return gnutls_record_send(conn->tls_session,s,slen);
  }
}

void peer_set_receiving_article(struct peer_connection *peer, article_t *article)
{
  if (peer->receiving_article) article_deref(peer->receiving_article);
  peer->receiving_article = article_ref(article);
}

void peer_set_message_id(struct peer_connection *peer, const char *message_id)
{
  if (peer->message_id) free(peer->message_id);
  peer->message_id = strdup(message_id);
}

struct peer_connection *peer_ref(struct peer_connection *peer)
{
  peer->refcount++;
  return peer;
}

void free_desired_range_set(struct range_interest *ranges, int num)
{
  if (ranges) {
    int x = num;
    while (x) {
      x--;
      free(ranges[x].group);
    }
    free(ranges);
  }
}

void peer_connection_free(struct peer_connection *peer)
{
  // XXX make sure that EVERYTHING attached to the peer is freed or dereferenced.
  if (peer->refcount<0) {
    fprintf(stderr,"bincache: BUG: peer reference count underflow.\n");
  }
  fprintf(stderr,"bincache: [debug %d]: peer_connection_free.\n",peer->num);

  if (peer->tls_session) {
    gnutls_deinit(peer->tls_session);
    peer->tls_session = NULL;
  }
  if (peer->tls_anoncred)
    gnutls_anon_free_client_credentials(peer->tls_anoncred);
  if (peer->tls_credentials)
    gnutls_certificate_free_credentials(peer->tls_credentials);
  if (peer->state != PEER_UNCONNECTED && peer->state != PEER_CLOSED) {
    peer_close(peer);
  }
  if (peer->fd>=0) {
    fprintf(stderr,"bincache: BUG: peer fd still open?\n");
    peer_close(peer);
  }
  if (peer->name) free(peer->name);
  if (peer->username) free(peer->username);
  if (peer->password) free(peer->password);
  free_desired_range_set(peer->ranges_desired_by_peer, peer->num_ranges_desired_by_peer);
  free_desired_range_set(peer->ranges_to_ask_for, peer->num_ranges_to_ask_for);
  free_desired_range_set(peer->ranges_asked_for, peer->num_ranges_asked_for);
  free(peer);
}

void peer_deref(struct peer_connection *peer)
{
  peer->refcount--;
  if (peer->refcount<=0) {
    peer_connection_free(peer);
  }
}

void article_note_asked(struct desired_article *article, struct peer_connection *peer)
{
  /* Add a mark to the desired article struct to the effect that a peer has
   * already been asked for that article.
   */
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  int x = article->numdonthave;
  while (x) {
    x--;
    if (realpeer==article->donthave[x]) {
      fprintf(stderr,"bincache: Error: %s added even though it's known not to be had.\n",article->message_id);
    }
  }
  x = article->numasked;
  while (x) {
    x--;
    if (realpeer==article->asked[x]) {
      fprintf(stderr,"bincache: Error: wanted to add article twice to %s\n",
        article->message_id);
    }
  }
  article->numasked ++;
  article->asked = realloc(article->asked,article->numasked*sizeof(struct peer_connection *));
  article->asked[article->numasked-1] = peer_ref(realpeer);
}

void note_high_fd(int fd)
{
  if (fd>=g_high_fd) g_high_fd = fd+1;
}

#ifdef WIN32
int inet_aton(const char *s, struct in_addr *ia)
{
  unsigned char addr[4];
  addr[0] = addr[1] = addr[2] = addr[3] = '\0';
  int x = 0;
  int y = 0;
  while (s[x]) {
    if (s[x]>='0' && s[x]<='9') {
      addr[y] *= 10;
      addr[y] += s[x]-'0';
    } else if (s[x]=='.') {
      y++;
      if (y>=4) return 0;
    } else {
      return 0;
    }
    x++;
  }
  ia->s_addr = (addr[0]<<24)|(addr[1]<<16)|(addr[2]<<8)|addr[3];
  return 1;
}
#endif

int resolve(struct sockaddr_in *si, const char *name)
{
  if (!name) {
    fprintf(stderr,"bincache: not trying to resolve nameless.\n");
    return 1;
  }
  struct hostent *host = gethostbyname(name);
  if (host && host->h_addr_list) {
    si->sin_addr.s_addr = *(unsigned int *)(host->h_addr_list[0]);
    return 1;
  } else if (inet_aton(name,&si->sin_addr)) {
    return 1;
  }
  fprintf(stderr,"bincache: error resolving %s.\n",name);
  return 0;
}

void peer_connect_unencrypted(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (resolve((struct sockaddr_in *)&master->conn, master->name)) {
    if (!master->conn.sin_port) master->conn.sin_port = htons(119);
    master->conn.sin_family = AF_INET;
    peer->fd = socket(PF_INET,SOCK_STREAM,0);
    if (peer->fd>=0) {
      if (!connect(peer->fd,(struct sockaddr *)&master->conn,sizeof(struct sockaddr_in))) {
        peer->state = PEER_READING_BANNER;
        FD_SET(peer->fd,&g_watched_rfds);
        note_high_fd(peer->fd);
      } else {
        peer_close(peer);
        fprintf(stderr,"bincache: peer_connect_unencrypted(): Error %d (%s) connecting to %s:%d.\n",sockerrno,sockstrerror(sockerrno),master->name,ntohs(master->conn.sin_port));
      }
    } else {
      fprintf(stderr,"bincache: socket error.\n");
    }
  }
}

void peer_set_tls_resume_watch(struct peer_connection *peer)
{
  if (gnutls_record_get_direction(peer->tls_session)) {
    FD_CLR(peer->fd,&g_watched_rfds);
    FD_SET(peer->fd,&g_watched_wfds);
  } else {
    FD_CLR(peer->fd,&g_watched_wfds);
    FD_SET(peer->fd,&g_watched_rfds);
  }
  note_high_fd(peer->fd);
}

void send_testpeer_for(struct peer_connection *peer)
{
  int x = g_num_cacheplex_connections;
  char testpeerbuf[1024];
  int wrote = snprintf(&testpeerbuf[0],1023,"TESTPEER %s:%d\015\012",peer->name,ntohs(peer->conn.sin_port));
  testpeerbuf[wrote]='\0';
  while (x) {
    x--;
    peer_ref(g_cacheplex_connections[x]);
    if (g_cacheplex_connections[x]->fd>=0) {
      peer_conn_writes(g_cacheplex_connections[x],&testpeerbuf[0]);
    }
    peer_deref(g_cacheplex_connections[x]);
  }
}

void peer_connect_tls_2(struct peer_connection *peer)
{
  int crttypes[2];
  crttypes[1] = 0;
  gnutls_init(&peer->tls_session,GNUTLS_CLIENT);
  gnutls_set_default_priority(peer->tls_session);
  if (peer->prot==PROTOCOL_CACHEPLEX_TRACKER) {
    crttypes[0] = GNUTLS_CRT_X509;
    if (peer->tls_server_cert && !peer->tls_credentials) {
      gnutls_certificate_allocate_credentials(&peer->tls_credentials);
      gnutls_certificate_set_x509_trust_file(peer->tls_credentials, peer->tls_server_cert, GNUTLS_X509_FMT_PEM);
      gnutls_credentials_set(peer->tls_session, GNUTLS_CRD_CERTIFICATE, peer->tls_credentials);
    }
    gnutls_certificate_type_set_priority(peer->tls_session, crttypes);
  } else if (peer->prot==PROTOCOL_BINCACHE_PEER) {
    crttypes[0] = GNUTLS_KX_ANON_DH;
    gnutls_anon_allocate_client_credentials(&peer->tls_anoncred);
    gnutls_credentials_set(peer->tls_session, GNUTLS_CRD_ANON, peer->tls_anoncred);
    gnutls_kx_set_priority(peer->tls_session, crttypes);
  } else {
    fprintf(stderr,"bincache: TLS connection on unsupported protocol.\n");
  }
  gnutls_transport_set_ptr(peer->tls_session, peer->fd); // Note, it whines about the typecast.
  gnutls_transport_set_lowat(peer->tls_session,0);
  int tlserr = gnutls_handshake(peer->tls_session);
  if (tlserr>=0) {
    if (peer->prot == PROTOCOL_BINCACHE_PEER) {
      peer_conn_writes(peer,"631 \015\012");
    }
    peer->state = PEER_READING_BANNER;
    FD_SET(peer->fd,&g_watched_rfds);
    note_high_fd(peer->fd);
    return;
  } else if ((tlserr==GNUTLS_E_INTERRUPTED) || (tlserr==GNUTLS_E_AGAIN)) {
    fprintf(stderr,"bincache: [debug]: peer_tls_connect_2: interrupted.\n");
    peer->state = PEER_TLS_HANDSHAKING;
    peer_set_tls_resume_watch(peer);
    return;
  }
  fprintf(stderr,"bincache: TLS handshake with %s failed.  Sending testpeer directive.\n",peer->name);
  send_testpeer_for(peer);
  gnutls_perror(tlserr);
  shutdown(peer->fd, SHUT_RDWR);
  gnutls_deinit(peer->tls_session);
  peer->tls_session = NULL;
}

int sockerrno_is_inprogress(int err)
{
  if (err==EINPROGRESS) return 1;
  #ifdef WIN32
  if (err==WSAEWOULDBLOCK) return 1;
  #endif
  return 0;
}

void peer_connect_tls(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (!resolve((struct sockaddr_in *)&master->conn, master->name)) {
    fprintf(stderr,"bincache: Error resolving %s\n",master->name);
    return;
  }
  fprintf(stderr,"[bincache]: debug: XXX connecting to %s:%d",master->name,ntohs(master->conn.sin_port));
  if (!master->conn.sin_port) master->conn.sin_port = htons(119);
  master->conn.sin_family = AF_INET;
  assert(peer->fd==-1);
  peer->tls = 1;
  peer->lastread = time(NULL);
  peer->fd = socket(PF_INET, SOCK_STREAM, 0);
  //fprintf(stderr,"XXX need to verify cert.\n");
  if (peer->fd>=0) {
    int one = 1;
    #ifndef WIN32
    if (ioctl(peer->fd,FIONBIO,&one)) {
    #else
    if (ioctlsocket(peer->fd,FIONBIO,&one)) {
    #endif
      fprintf(stderr,"bincache: Error setting TLS socket nonblocking.\n");
      close(peer->fd);
      peer->fd = -1;
      return;
    }
    if (peer->prot == PROTOCOL_CACHEPLEX_TRACKER) {
      struct sockaddr bindaddr = g_bincache_listen_addr;
      ((struct sockaddr_in *)&bindaddr)->sin_addr.s_addr = INADDR_ANY;
      if (bind(peer->fd, &bindaddr,sizeof(struct sockaddr_in))) {
        fprintf(stderr,"bincache: bind() error %d (%s)\n",errno,strerror(errno));
      }
    }
    if (!connect(peer->fd,(struct sockaddr *)&master->conn,sizeof(struct sockaddr_in))) {
      peer_connect_tls_2(peer);
    } else if (sockerrno_is_inprogress(sockerrno)) {
      peer->state = PEER_TCP_CONNECTING;
      FD_SET(peer->fd,&g_watched_wfds);
      note_high_fd(peer->fd);
      fprintf(stderr,"bincache: connect() in progress.\n");
      return;
    } else {
      fprintf(stderr,"bincache: connect to SSL peer failed: %s.\n",sockstrerror(sockerrno));
    }
    close(peer->fd);
    peer->fd = -1;
  } else {
    fprintf(stderr,"bincache: socket() failed.\n");
  }
}

void peer_connect_apropos(struct peer_connection *peer)
{
  if (peer->tls) {
    peer_connect_tls(peer);
  } else {
    peer_connect_unencrypted(peer);
  }
}

void note_command_issued_time(struct peer_connection *peer)
{
  #ifdef WIN32
  fprintf(stderr,"bincache: BUG: XXX highres time on Windows not yet supported.\n");
  #else
  gettimeofday(&peer->lastcmdtime,NULL);
  #endif
}

void query_peer_for_head(struct peer_connection *peer, struct desired_article *article)
{
  char buf[512];
  snprintf(&buf[0],512,"HEAD <%s>\015\012",article->message_id);
  buf[511] = '\0';
  switch (peer->state) {
    case PEER_UNCONNECTED:
    case PEER_CLOSED:
      peer_connect_apropos(peer);
      break;
    case PEER_BIDI_READY_FOR_COMMAND:
    case PEER_READY_FOR_COMMAND:
      if (peer->trace>1) fprintf(stderr,"bincache: [Debug]: %s: req head %s\n",peer->name,article->message_id);
      peer->state = PEER_WAITING_FOR_HEAD_RESP;
      peer_ref(peer);
      peer_conn_writes(peer,&buf[0]);
      if (peer->state==PEER_WAITING_FOR_HEAD_RESP) {
        peer_set_message_id(peer,article->message_id);
        article_note_asked(article,peer);
      }
      peer_deref(peer);
      break;
    default:
      fprintf(stderr,"bincache: %s: invalid peer state %d, wanting to request head.\n",peer->name,peer->state);
      breakable();
  }
}

void query_peer_for_newsgroups_header(struct peer_connection *peer, struct desired_article *article)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  switch (peer->state) {
    case PEER_UNCONNECTED:
    case PEER_CLOSED:
      peer_connect_apropos(peer);
      break;
    case PEER_READY_FOR_COMMAND:
    case PEER_BIDI_READY_FOR_COMMAND:
      if (master->hdr_not_supported || master->hdr_newsgroups_not_supported) {
        /* Falling back to XHDR may be pointless; I don't know if any wild-type
         * implementations support XHDR on message IDs.  So we just fall back to
         * HEAD.
         */
        query_peer_for_head(peer, article);
        return;
      }
      char buf[512];
      note_command_issued_time(peer);
      snprintf(&buf[0],512,"HDR Newsgroups <%s>\015\012",article->message_id);
      buf[511] = '\0';
      peer->state = PEER_WAITING_FOR_NEWSGROUPS_HDR;
      peer_ref(peer);
      peer_conn_writes(peer,&buf[0]);
      if (peer->state==PEER_WAITING_FOR_NEWSGROUPS_HDR) {
        peer_set_message_id(peer,article->message_id);
        article_note_asked(article,peer);
      }
      peer_deref(peer);
      break;
    default:
      /* No need to whine, since this is going to be called in states when the 
       * peer is busy. */
      /* fprintf(stderr,"bincache: %s: invalid peer state %d\n",peer->name, peer->state);*/
      return;
  }
}

void query_peer_for_date_header(struct peer_connection *peer, struct desired_article *article)
{
  char buf[512];
  snprintf(&buf[0],512,"HDR Date <%s>\015\012",article->message_id);
  buf[511] = '\0';
  switch (peer->state) {
    case PEER_READY_FOR_COMMAND:
    case PEER_BIDI_READY_FOR_COMMAND:
      peer->state = PEER_WAITING_FOR_DATE_HDR;
      peer_ref(peer);
      peer_conn_writes(peer,&buf[0]);
      if (peer->state==PEER_WAITING_FOR_DATE_HDR) {
        peer_set_message_id(peer,article->message_id);
        article_note_asked(article,peer);
      }
      peer_deref(peer);
      break;
    default:
      fprintf(stderr,"bincache: querying %s for date header: invalid peer state %d\n",peer->name,peer->state);
      breakable();
  }
}

void peer_watch_for_write(struct peer_connection *peer)
{
  if (peer->fd>=0) {
    FD_SET(peer->fd,&g_watched_wfds);
    if (peer->prot!=PROTOCOL_BINCACHE_PEER && peer->prot!=PROTOCOL_CACHEPLEX_TRACKER) {
      FD_CLR(peer->fd,&g_watched_rfds);
    }
  } else {
    fprintf(stderr, "bincache: [bug]: Trying to watch an invalid file descriptor.\n");
    breakable();
  }
}

void peer_watch_for_read(struct peer_connection *peer)
{
  if (peer->fd>=0) {
    FD_CLR(peer->fd,&g_watched_wfds);
    FD_SET(peer->fd,&g_watched_rfds);
  } else {
    fprintf(stderr, "bincache: [bug]: Trying to watch an invalid file descriptor.\n");
    breakable();
  }
}

void query_peer_for_stat(struct peer_connection *peer, struct desired_article *article)
{
  char buf[512];
  snprintf(&buf[0],512,"STAT <%s>\015\012",article->message_id);
  buf[511]='\0';
  switch (peer->state) {
    case PEER_READY_FOR_COMMAND:
    case PEER_BIDI_READY_FOR_COMMAND:
      peer->state = PEER_WAITING_FOR_STAT_RESPONSE;
      peer_ref(peer);
      peer_conn_writes(peer,&buf[0]);
      if (peer->state==PEER_WAITING_FOR_STAT_RESPONSE) {
        peer_watch_for_read(peer);
        peer_set_message_id(peer,article->message_id);
        article_note_asked(article,peer);
      }
      peer_deref(peer);
      break;
    default:
      fprintf(stderr,"bincache: %s invalid peer state %d statting\n",peer->name,peer->state);
      breakable();
  }
}

/* Return 1 if:
 * The peer is waiting for a response to a request for this article.
 * The peer has already said it has this article.
 * The peer has already said it doesn't have this article.
 */
int article_queried_from(struct desired_article *article, struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  int x = article->numasked;
  if (article->numasked<0) {
    fprintf(stderr,"bincache: BUG: <0 peers asked.\n");
  }
  while (x) {
    x--;
    if (article->asked[x]==master) return 1;
  }
  x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]==master) return 1;
  }
  x = article->numdonthave;
  while (x) {
    x--;
    if (article->donthave[x]==master) return 1;
  }
  return 0;
}

char unescape(const char *escaped)
{
  if (*escaped) {
    return ((escaped[0]-'@')<<4)|(escaped[1]-'@');
  }
  return '\0';
}

char *cachename_to_messageid(const char *c1, const char *c2, const char *c3)
{
  int c3len = strlen(c3);
  if (c3len>8) c3len-=8;
  char *ret;
  ret = malloc(c3len+3);
  if (*c1=='^') {
    ret[0] = unescape(&c1[1]);
  } else {
    ret[0] = *c1;
  }
  if (*c2=='^') {
    ret[1] = unescape(&c2[1]);
  } else {
    ret[1] = *c2;
  }
  int x = 0;
  int y = 2;
  while (x<c3len) {
    if (c3[x]!='^') {
      ret[y++] = c3[x];
    } else {
      ret[y++] = unescape(&c3[x+1]);
      x+=2;
    }
    x++;
  }
  ret[y] = '\0';
  return ret;
}

void write_safe_char(char c, char *dest, int *dptr)
{
  if ((c>='0' && c<='9') || (c>='@' && c<='Z') || (c>='a' && c<='z')) {
    dest[*dptr] = c;
    ++(*dptr);
  } else {
    dest[*dptr] = '^';
    ++(*dptr);
    dest[*dptr] = (c>>4)+'@';
    ++(*dptr);
    dest[*dptr] = (c&0xf)+'@';
    ++(*dptr);
  }
}

const char *message_id_to_path_stem(const char *message_id)
{
  int y = 0;
  static char path_stem[PATH_MAX];
  char dir[PATH_MAX];
  if (message_id[0]) {
    write_safe_char(message_id[0],&path_stem[0],&y);
    path_stem[y++] = '/';
    path_stem[y]='\0';
    snprintf(&dir[0],PATH_MAX,"%s/%s",g_cache_dir,&path_stem[0]);
    dir[PATH_MAX-1]='\0';
    makedir(&dir[0]);
  }
  if (message_id[1]) {
    write_safe_char(message_id[1],&path_stem[0],&y);
    path_stem[y++] = '/';
    path_stem[y]='\0';
    snprintf(&dir[0],PATH_MAX,"%s/%s",g_cache_dir,&path_stem[0]);
    dir[PATH_MAX-1]='\0';
    makedir(&dir[0]);
  }
  int x = 2;
  while (message_id[x] && y+4<PATH_MAX) {
    write_safe_char(message_id[x],&path_stem[0],&y);
    x++;
  }
  path_stem[y++] = '\0';
  return &path_stem[0];
}

void article_create_articlepart_file(struct desired_article *article)
{
  char headerfile[PATH_MAX];
  char partfile[PATH_MAX];
  snprintf(&headerfile[0],PATH_MAX,"%s/%s.head",g_cache_dir,message_id_to_path_stem(article->message_id));
  headerfile[PATH_MAX-1] = '\0';
  snprintf(&partfile[0],PATH_MAX,"%s/%s.articlepart",g_cache_dir,message_id_to_path_stem(article->message_id));
  if (!rename(&headerfile[0],&partfile[0])) {
    if (article->articlefd>=0) close(article->articlefd);
    article->articlefd = open(&partfile[0],O_BINARY|O_RDWR);
    if (article->articlefd>=0) {
      article->articlefd_wpos = lseek(article->articlefd,0,SEEK_END);
      if (write(article->articlefd,"\015\012",2)==2) {
        article->articlefd_wpos += 2;
        /*if (write(article->articlefd,"\015\012",2)!=2) {
          fprintf(stderr,"bincache: write() error %d (%s)\n",errno, strerror(errno));
        } else {
	  article->articlefd_wpos += 2;
	}*/
      } else {
        fprintf(stderr,"bincache: write() error %d (%s)\n",errno, strerror(errno));
      }
    } else {
      fprintf(stderr,"bincache: open() error.\n");
    }
  } else {
    fprintf(stderr,"bincache: error %d (%s) renaming %s to %s\n", errno, strerror(errno), &headerfile[0], &partfile[0]);
  }
}

void article_open_body_file(struct desired_article *article)
{
  if (article->articlefd>=0) close(article->articlefd);
  char buf[PATH_MAX];
  snprintf(&buf[0],PATH_MAX,"%s/%s.bodypart", g_cache_dir, message_id_to_path_stem(article->message_id));
  buf[PATH_MAX-1]='\0';
  article->articlefd = open(&buf[0],O_RDWR|O_BINARY|O_CREAT,0666);
  article->articlefd_wpos = 0;
}

void query_peer_for_body(struct peer_connection *peer, struct desired_article *article)
{
  char buf[512];
  switch (peer->state) {
    case PEER_UNCONNECTED:
      peer_connect_apropos(peer);
      break;
    case PEER_READY_FOR_COMMAND:
    case PEER_BIDI_READY_FOR_COMMAND:
      if (article->bodycomplete) {
        fprintf(stderr,"bincache: [BUG]: trying to acquire a message with a complete body.\n");
	breakable();
	return;
      }
      fprintf(stderr,"bincache: %llx:%s BODY %s\n",(long long unsigned int)peer,peer->name,article->message_id);
      snprintf(&buf[0],512,"BODY <%s>\015\012",article->message_id);
      buf[511] = '\0';
      if (!article->headerscomplete) {
        article_open_body_file(article);
      } else {
        article_create_articlepart_file(article);
      }
      peer->state = PEER_WAITING_FOR_BODY_RESP;
      peer_ref(peer);
      peer_conn_writes(peer,&buf[0]);
      note_command_issued_time(peer);
      if (peer->state == PEER_WAITING_FOR_BODY_RESP) {
        article->gettingbody = 1;
        peer_set_message_id(peer, article->message_id);
        if (!article_queried_from(article, peer)) {
          article_note_asked(article, peer);
        }
      }
      break;
    case PEER_TCP_CONNECTING:
    case PEER_TLS_HANDSHAKING:
      break;
    default:
      fprintf(stderr,"bincache: asking for body in peer state %d\n", peer->state);
      breakable();
  }
}

void query_peer_for(struct peer_connection *peer, struct desired_article *article)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (article_queried_from(article,master)) return;
  if (peer->prot==PROTOCOL_BINCACHE_PEER && peer->state==PEER_BIDI_READY_FOR_COMMAND) {
    /* Assuming all Bincache connections are relevant to all message-IDs. 
       This will waste some bandwidth, but hopefully not too much.
     */
    query_peer_for_stat(peer,article);
    return;
  }
  if (!article->newsgroups_header) {
    if (peer->isfullfeed) {
      query_peer_for_newsgroups_header(peer,article);
    }
    return;
    // TODO query peers based on relevancy to clients, as well
  }
  if (!article->dateparsed) {
    if (master->isfullfeed) {
      query_peer_for_date_header(peer,article);
    }
    return;
    // TODO query peers known to be relevant to that group, as well
  }
  // XXX M2 peers relevant to group/date should be asked if they have the
  // article.
}

void ask_all_peers(struct desired_article *article)
{
  if (!g_num_peer_connections) {
    article->unavailable = 1;
    return;
  }
  int x = g_num_peer_connections;
  //char buf[513];
  //snprintf(&buf[0],512,"STAT <%s>\015\012",article->message_id);
  //buf[512]='\0';
  while (x) {
    x--;
    fprintf(stderr,"bincache: debug: Querying %d for %s\n",x,article->message_id);
    query_peer_for(g_peer_connections[x],article);
  }
}

void client_close(struct client_connection *conn)
{
  if (conn->fd>=0) {
    close(conn->fd);
    FD_CLR(conn->fd,&g_watched_rfds);
    FD_CLR(conn->fd,&g_watched_wfds);
    conn->fd = -1;
  }
  conn->state = CLIENT_CLOSED;
}

void client_conn_writes(struct client_connection *conn, const char *s)
{
  if (writes(conn->fd,s)) return;
  client_close(conn);
}

struct desired_article *desired_article_alloc(const char *message_id)
{
  struct desired_article *ret = calloc(1, sizeof(struct desired_article));
  ret->articlefd = -1;
  ret->message_id = strdup(message_id);
  ret->last_activity = time(NULL);
  return ret;
}

void article_attach_leecher(struct desired_article *article, struct client_connection *client)
{
  ++article->numleechers;
  article->leechers = realloc(article->leechers,sizeof(struct client_connection *)*article->numleechers);
  article->leechers[article->numleechers-1] = client_ref(client);
  if (client->article) {
    fprintf(stderr,"bincache: error: replacing leecher for article with leecher set.\n");
  }
  client->article = article_ref(article);
}

struct desired_article *desire_article(struct client_connection *client, const char *message_id)
{
  int midlen = strlen(message_id);
  int x = g_num_desired_articles;
  struct desired_article *article;
  while (x) {
    x--;
    if (!memcmp(message_id,g_desired_articles[x]->message_id,midlen)) {
      article = g_desired_articles[x];
      article->last_activity = time(NULL);
      if (article->unavailable) {
        client_conn_writes(client,"430 no peers have it (yet?) (no peers connected?)\015\012");
        return NULL;
      }
      article_attach_leecher(article,client);
      ask_all_peers(article);
      return NULL;
    }
  }
  article = desired_article_alloc(message_id);
  ++g_num_desired_articles;
  g_desired_articles = realloc(g_desired_articles,g_num_desired_articles*sizeof(struct desired_article *));
  g_desired_articles[g_num_desired_articles-1] = article;
  article_attach_leecher(article,client);
  ask_all_peers(article);
  if (article->unavailable) {
    client_conn_writes(client,"430 no peers have it (yet?) (no peers connected?)\015\012");
  }
  return article;
}

int ipaddr_a(struct sockaddr *sa)
{
  return ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr)>>24;
}

int ipaddr_b(struct sockaddr *sa) 
{
  return (ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr)>>16)&0xff;
}

int ipaddr_c(struct sockaddr *sa)
{
  return (ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr)>>8)&0xff;
}

int ipaddr_d(struct sockaddr *sa) 
{
  return ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr)&0xff;
}

int ipport(struct sockaddr *sa)
{
  return ntohs(((struct sockaddr_in *)sa)->sin_port);
}

protocol_enum_type protocol_by_name(const char *s)
{
  if (!memcmp(s,"cacheplex",9)) return PROTOCOL_CACHEPLEX_TRACKER;
  if (!memcmp(s,"bincache",8)) return PROTOCOL_BINCACHE_PEER;
  fprintf(stderr,"bincache: Fatal error: unknown protocol %s\n",s);
  exit(1);
}

struct peer_connection *slave_peer(struct peer_connection *master)
{
  struct peer_connection *peer = calloc(1, sizeof(struct peer_connection));
  peer->master = master;
  peer->prot = master->prot;
  peer->headers_only = master->headers_only;
  peer->trace = master->trace;
  peer->isfullfeed = master->isfullfeed;
  peer->num = peer_serial_number++;
  return peer;
}

void add_slaves(struct peer_connection *current_server)
{
  if (!current_server) return;
  if (!current_server->numslaves) return;
  g_num_slave_connections += current_server->numslaves;
  int i = g_num_peer_connections;
  g_num_peer_connections += current_server->numslaves;
  g_peer_connections = realloc(g_peer_connections,g_num_peer_connections*sizeof(struct peer_connection *));
  int slaves = current_server->numslaves;
  while (slaves) {
    g_peer_connections[i] = peer_ref(slave_peer(current_server));
    i ++;
    slaves --;
  }
}

void read_ini_file(void)
{
  FILE *f = fopen("bincache.ini","r");
  struct peer_connection *current_server = NULL;
  char buf[8192];
  if (!f) {
    fprintf(stderr,"bincache: Error opening bincache.ini.\n");
    return;
  }
  char *str;
  while ((str=fgets(&buf[0],8192,f))) {
    if (buf[0]=='[') {
      str = index(str,']');
      if (str) {
        *str = '\0';
        if (!memcmp(&buf[1],"default",7)) {
          current_server = g_default_server;
        } else if (!memcmp(&buf[1],"group ",6)) {
          printf("bincache: group settings not yet implemented.\n");
        } else {
          add_slaves(current_server);
          ++g_num_peer_connections;
          g_peer_connections = realloc(g_peer_connections,g_num_peer_connections*sizeof(struct peer_connection *));
          if (current_server) {
            peer_deref(current_server);
          }
          current_server = peer_ref(calloc(1,sizeof(struct peer_connection)));
          current_server->fd = -1;
          current_server->num = peer_serial_number++;
          g_peer_connections[g_num_peer_connections-1] = peer_ref(current_server);
          current_server->name = strdup(&buf[1]);
        }
      } else {
        fprintf(stderr,"bincache: Syntax error in ini file: ] expected.\n");
        fclose(f);
        return;
      }
    } else if (!memcmp(&buf[0],"bincache-listen-address=",24)) {
      if (!resolve((struct sockaddr_in *)&g_bincache_listen_addr,&buf[24])) {
        fprintf(stderr,"bincache: error parsing bincache protocol listening address.\n");
      }
    } else if (!memcmp(&buf[0],"bincache-listen-port=",21)) {
      g_bincache_listen_port = atoi(&buf[21]);
    } else if (!memcmp(&buf[0],"cert=",5)) {
      if (current_server->tls_server_cert) 
        free(current_server->tls_server_cert);
      current_server->tls_server_cert = strdup(&buf[5]);
    } else if (!memcmp(&buf[0],"connections=",12)) {
      current_server->numslaves = atoi(&buf[12])-1;
    } else if (!memcmp(&buf[0],"fullfeed=",9)) {
      current_server->isfullfeed = atoi(&buf[9]);
    } else if (!memcmp(&buf[0],"headersonly=",12)) {
      if (buf[12]!='0') {
        current_server->headers_only = 1;
      } else {
        current_server->headers_only = 0;
      }
    } else if (!memcmp(&buf[0],"indexmaster=",12)) {
      printf("bincache: group settings not yet implemented.\n");
    } else if (!memcmp(&buf[0],"nntp-listen-address=",20)) {
      if ((str=index(&buf[20],'\015'))) {
        *str = '\0';
      } else if ((str=index(&buf[20],'\012'))) {
        *str = '\0';
      }
      if (!resolve((struct sockaddr_in *)&g_nntp_listen_addr,&buf[20])) {
        fprintf(stderr,"bincache: error parsing NNTP listening address.\n");
      } else {
        printf("bincache: will listen on %u.%u.%u.%u\n", 
          ipaddr_a(&g_nntp_listen_addr), ipaddr_b(&g_nntp_listen_addr),
          ipaddr_c(&g_nntp_listen_addr), ipaddr_d(&g_nntp_listen_addr));
      }
    } else if (!memcmp(&buf[0],"nntp-listen-port=",17)) {
      int port = atoi(&buf[17]);
      if (port>=0 && port<=65535) {
        ((struct sockaddr_in *)&g_nntp_listen_addr)->sin_port = htons(port);
      } else {
        fprintf(stderr,"bincache: invalid nntp-listen-port %d specified, must be in range 1-65535\n", port);
      }
    } else if (!memcmp(&buf[0],"password=",9)) {
      if (current_server->password) free(current_server->password);
      int passwordlen = strlen(&buf[9]);
      current_server->password = malloc(passwordlen+1);
      memcpy(current_server->password,&buf[9],passwordlen);
      if (passwordlen && current_server->password[passwordlen-1]=='\012')
        current_server->password[passwordlen-1] = '\0';
    } else if (!memcmp(&buf[0],"port=",5)) {
      current_server->conn.sin_port = htons(atoi(&buf[5]));
    } else if (!memcmp(&buf[0],"protocol=",9)) {
      current_server->prot = protocol_by_name(&buf[9]);
      if (current_server->prot == PROTOCOL_CACHEPLEX_TRACKER) {
        g_num_cacheplex_connections++;
        g_cacheplex_connections = realloc(g_cacheplex_connections, g_num_cacheplex_connections * sizeof(struct peer_connection *));
        g_cacheplex_connections[g_num_cacheplex_connections-1] = current_server;
        g_num_peer_connections--;
      }
    } else if (!memcmp(&buf[0],"trace=",6)) {
      current_server->trace = atoi(&buf[6]);
    } else if (!memcmp(&buf[0],"user=",5)) {
      if (current_server->username) free(current_server->username);
      int usernamelen = strlen(&buf[5]);
      current_server->username = malloc(usernamelen+1);
      memcpy(current_server->username,&buf[5],usernamelen);
      current_server->username[usernamelen]='\0';
      if (usernamelen && current_server->username[usernamelen-1]=='\012')
        current_server->username[usernamelen-1] = '\0';
    } else if (buf[0]!='\012' && buf[0]!='#') {
      fprintf(stderr,"Unexpected command %s in ini file.\n",&buf[0]);
    }
  }
  fclose(f);
  if (current_server) add_slaves(current_server);
}

void init_regexps(void)
{
  if (regcomp(&g_date_wdy_DMYHMS_Z, "^Date: [A-Z][a-z]+, ([0-9]+) ([A-Z][a-z]+) ([0-9]+) ([0-9]+):([0-9]+):([0-9]+) ([-+0-9A-Z]+)\015", REG_EXTENDED)) {
    fprintf(stderr,"bincache: Error compiling regex.\n");
    exit(-1);
  }
  // Date: 30 Aug 2007 07:41:21 GMT XXX
  if (regcomp(&g_date_DMYHMS_z, "^Date: ([0-9]+) ([A-Z][a-z]+) ([0-9]+) ([0-9]+):([0-9]+):([0-9]+) ([A-Z]+)\015", REG_EXTENDED)) {
    fprintf(stderr,"bincache: Error compiling regex.\n");
    exit(-1);
  }
}

void init_crypto(void)
{
  gnutls_global_init();
//  gnutls_certificate_allocate_credentials(&g_x509_credentials);
//  gnutls_certificate_set_x509_trust_file(g_x509_credentials, "cacheplex.cert", GNUTLS_X509_FMT_PEM);
}

void regenerate_broken_time(struct brokentime *btime)
{
  struct tm *ptm;

  ptm = gmtime(&btime->stamp);
  if (ptm) {
    btime->year = ptm->tm_year+1900;
    btime->month = ptm->tm_mon+1;
    btime->day = ptm->tm_mday;
    btime->hour = ptm->tm_hour;
    btime->minute = ptm->tm_min;
  } else {
    fprintf(stderr,"bincache: gmtime_r failed.\n");
  }
}

void init_bincache_listener(void)
{
#ifndef WIN32
  if (setenv("TZ","GMT",1)) {
    fprintf(stderr,"bincache: Error setting TZ to GMT.\n");
  }
#endif
  if (g_bincache_listen_port>=0) {
    /* We only generate DH params if we will be using DHE. */
    gnutls_anon_allocate_server_credentials(&g_anon_credentials);
    gnutls_dh_params_init(&g_dh_params);
    gnutls_dh_params_generate2(g_dh_params, 1024);
    gnutls_anon_set_server_dh_params(g_anon_credentials, g_dh_params);
    /* TODO regenerate DH params every so often.  Probably daily or weekly. */
    struct sockaddr_in *bincache_sin = (struct sockaddr_in *)&g_bincache_listen_addr;
    bincache_sin->sin_port = htons(g_bincache_listen_port);
    g_bincache_listen_socket = socket(PF_INET, SOCK_STREAM, 0);
    if (g_bincache_listen_socket<0) {
      fprintf(stderr,"bincache: Error creating peer listening socket.\n");
      exit(-1);
    }
    int one = 1;
    if (setsockopt(g_bincache_listen_socket, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(int))) {
      fprintf(stderr,"Error setting SO_REUSEADDR on socket.\n");
      exit(-1);
    }
    bincache_sin->sin_family = AF_INET;
    if (bind(g_bincache_listen_socket, &g_bincache_listen_addr,sizeof(struct sockaddr_in))) {
      fprintf(stderr,"bincache: error %d (%s) binding listen socket to port %d.\n",errno, strerror(errno),g_bincache_listen_port);
      exit(-1);
    }
    if (!listen(g_bincache_listen_socket, 5)) {
      struct sockaddr_in sin;
      socklen_t sinl = sizeof(sin);
      if (getsockname(g_bincache_listen_socket,(struct sockaddr *)&sin,&sinl)) {
        fprintf(stderr,"bincache: getsockname() error.\n");
      } else {
        g_bincache_real_listen_port = ntohs(sin.sin_port);
        fprintf(stdout,"bincache: Listening for bincache peers on %d\n",g_bincache_real_listen_port);
      }
    }
    if (g_bincache_listen_socket>=g_high_fd) g_high_fd = g_bincache_listen_socket+1;
    FD_SET(g_bincache_listen_socket,&g_watched_rfds);
  }
}

int monthnum (const char *mon)
{
  if (!memcmp(mon,"Jan",3)) return 1;
  if (!memcmp(mon,"Feb",3)) return 2;
  if (!memcmp(mon,"Mar",3)) return 3;
  if (!memcmp(mon,"Apr",3)) return 4;
  if (!memcmp(mon,"May",3)) return 5;
  if (!memcmp(mon,"Jun",3)) return 6;
  if (!memcmp(mon,"Jul",3)) return 7;
  if (!memcmp(mon,"Aug",3)) return 8;
  if (!memcmp(mon,"Sep",3)) return 9;
  if (!memcmp(mon,"Oct",3)) return 10;
  if (!memcmp(mon,"Nov",3)) return 11;
  if (!memcmp(mon,"Dec",3)) return 12;
  return 0;
}

void article_set_date(struct desired_article *article, const char *date)
{
  // XXX set date.
  // Date: Fri, 20 Apr 2007 00:39:24 +0200
  regmatch_t datematch[8];
  if ((!regexec(&g_date_wdy_DMYHMS_Z, date, 8, &datematch[0], 0)) || 
    (!regexec(&g_date_DMYHMS_z, date, 8, &datematch[0], 0))) {
    // 0 - whole thing.
    if (datematch[1].rm_so>=0) article->day = atoi(&date[datematch[1].rm_so]);
    if (datematch[2].rm_so>=0) 
      article->month = monthnum(&date[datematch[2].rm_so]);
    if (datematch[3].rm_so>=0) article->year = atoi(&date[datematch[3].rm_so]);
    if (datematch[4].rm_so>=0) article->hour = atoi(&date[datematch[4].rm_so]);
    if (datematch[5].rm_so>=0) article->minute = atoi(&date[datematch[5].rm_so]);
    static struct tm tm;
    tm.tm_min  = article->minute;
    tm.tm_hour = article->hour;
    tm.tm_mday = article->day;
    tm.tm_mon  = article->month-1;
    tm.tm_year = article->year - 1900;
    int zonelen = datematch[7].rm_eo-datematch[7].rm_so;
    char *zone = malloc(zonelen+1);
    memcpy(zone,&date[datematch[7].rm_so],zonelen);
    zone[zonelen] = '\0';
    #ifndef WIN32
    tm.tm_zone = zone;
    #endif
    article->timestamp = mktime(&tm);
    free(zone);
    if (article->timestamp==-1) {
      fprintf(stderr,"bincache: Article Time -1\n");
    }
    article->dateparsed = 1;
    // 6 - seconds.  Ignore.
    // 7 - Timezone.  XXX IIRC, one of the RFCs authorizes a server to change
    //   date headers...
  } else {
    fprintf(stderr, "bincache: Regex mismatch: cannot parse '%s'\n",date);
  }
  /*
  #ifndef WIN32
  struct tm tm;
  memset(&tm,'\0',sizeof(tm));
  char *dateend = strptime(date, "Date: %a,%n%d%n%b%n%Y%n%T%n%z", &tm);
  if (!dateend) {
    dateend = strptime(date, "Date: %a,%n%d%n%b%n%Y%n%T%n%Z", &tm);
  }
  if (dateend) {
    if (*dateend=='\015') {
      article->date = tm;
      article->dateparsed = 1;
    } else {
      fprintf(stderr,"bincache: BUG: unhandled Usenet date format in date %s\n",date);
    }
  } else {
    fprintf(stderr,"bincache: XXX BUG: date format not supported.\n");
  }
  #else
  fprintf(stderr,"bincache: XXX BUG: date not supported.\n");
  #endif
  */
}

void note_posessed(struct desired_article *article)
{
  g_posessed_articles = realloc(g_posessed_articles,sizeof(struct desired_article *)*(g_num_posessed_articles+1));
  g_posessed_articles[g_num_posessed_articles++] = article_ref(article);
}

void load_cached_article_data(const char *name1, const char *name2, const char *name3)
{
  char buf[1025];
  snprintf(&buf[0],1024,"%s/%s/%s/%s",g_cache_dir,name1,name2,name3);
  FILE *f = fopen(&buf[0],"rb");
  if (!f) return;
  char *s;
  struct desired_article *article = desired_article_alloc(s=cachename_to_messageid(name1,name2,name3));
  free(s);
  while (fgets(&buf[0],1025,f)) {
    if (buf[0]=='\015') {
      fseek(f,0,SEEK_END);
    } else {
      buf[1024] = '\0';
      int buflen = strlen(&buf[0]);
      if (buf[buflen]=='\012' && buflen>0) buf[--buflen] = '\0';
      if (buf[buflen]=='\015' && buflen>0) buf[--buflen] = '\0';
      if (!memcmp(&buf[0],"Newsgroups: ",12)) {
        if (!article->newsgroups_header) {
          article->newsgroups_header = malloc(buflen-11);
          memcpy(&article->newsgroups_header[0],&buf[12],buflen-14);
          article->newsgroups_header[buflen-14] = '\0';
          article->newsgroups_header_len = buflen-14;
          //fprintf(stderr,"bincache: [bedug]: scanning directory found article in groups %s\n",article->newsgroups_header);
        } else {
          fprintf(stderr,"bincache: malformed article has multiple Newsgroups: headers.\n");
        }
      } else if (!memcmp(&buf[0],"Date: ",6)) {
        fprintf(stderr,"bincache: [debug]: scanning directory found article with date %s\n",&buf[0]);
        article_set_date(article,&buf[0]);
      }
      /* Continuation lines not supported.   XXX. */
    }
  }
  if (article->newsgroups_header && article->dateparsed) {
    note_posessed(article);
    /* Mark the article as complete. */
    article->headerscomplete = 1;
    article->bodycomplete = 1;
  }
  fclose(f);
}

unsigned long stamptorange(time_t t)
{
  struct tm *tm;
  if ((tm=gmtime(&t))) {
    return (tm->tm_year*1000000)+(tm->tm_mon*10000)+(tm->tm_mday*100)+tm->tm_hour+1900010000;
  } else {
    fprintf(stderr,"bincache: gmtime_r failed.\n");
    return 0;
  }
}

void add_article_to_posessed_ranges(struct desired_article *article)
{
  int x = g_num_posessed_ranges;
  unsigned long overlapmin = stamptorange(article->timestamp-3600);
  unsigned long overlapmax = stamptorange(article->timestamp+3600);
  unsigned long articlet = stamptorange(article->timestamp);
  while (x) {
    x--;
    /* Possibilities: */
    /* article max < range min: don't include. */
    /* article min > range max: don't include. */
    /* otherwise, if group matches, include. */
    /* */
    if (overlapmax>=g_posessed_ranges[x].minl && overlapmin<=g_posessed_ranges[x].maxl) {
      /* XXX does not really handle crossposts properly. */
      if (!strcmp(g_posessed_ranges[x].group, article->newsgroups_header)) {
        if (articlet<g_posessed_ranges[x].minl) {
          g_posessed_ranges[x].minl = articlet;
        } else if (articlet>g_posessed_ranges[x].maxl) {
          g_posessed_ranges[x].maxl = articlet;
        }
        return;
      }
    }
  }
  int r = g_num_posessed_ranges;
  g_num_posessed_ranges++;
  g_posessed_ranges = realloc(g_posessed_ranges,g_num_posessed_ranges * sizeof(struct range_interest));
  g_posessed_ranges[r].group = strdup(article->newsgroups_header);
  g_posessed_ranges[r].minl = articlet;
  g_posessed_ranges[r].maxl = articlet;
  g_posessed_ranges[r].last_activity = 0;
}

void compact_posessed_range_list(void)
{
  int x = g_num_posessed_ranges;
  while (x) {
    x--;
    int y = x;
    int minstamp = g_posessed_ranges[x].minl;
    int maxstamp = g_posessed_ranges[x].maxl;
    while (y) {
      y--;
      if (g_posessed_ranges[y].maxl>minstamp && 
        g_posessed_ranges[y].minl<maxstamp) {
        if (!strcmp(g_posessed_ranges[x].group,g_posessed_ranges[y].group)) {
	  if (g_posessed_ranges[x].minl<g_posessed_ranges[y].minl) {
	    /* X is less than Y. */
	    g_posessed_ranges[y].minl = g_posessed_ranges[x].minl;
	  }
	  if (g_posessed_ranges[x].maxl>g_posessed_ranges[y].maxl) {
	    g_posessed_ranges[y].maxl = g_posessed_ranges[x].maxl;
          }
          free(g_posessed_ranges[x].group);
          g_num_posessed_ranges--;
          g_posessed_ranges[x] = g_posessed_ranges[g_num_posessed_ranges];
        }
      }
    }
  }
}

void init_posessed_range_list(void)
{
  int x = g_num_posessed_articles;
  while (x) {
    x--;
    add_article_to_posessed_ranges(g_posessed_articles[x]);
  }
  compact_posessed_range_list();
}

int buf_write_range(char *buf, int *bufpos, int buflen, int min, int max)
{
  if (22+*bufpos>=buflen) return 0;
  int len = snprintf(&buf[*bufpos],buflen-*bufpos,"%d",min);
  *bufpos += len;
  if (min!=max) {
    if (23+*bufpos>=buflen) return 0;
    buf[*bufpos] = '-';
    ++*bufpos;
    len = snprintf(&buf[*bufpos],buflen-*bufpos,"%d",max);
    *bufpos += len;
  }
  return 1;
}


int write_haverange(char *buf, int buflen,struct range_interest *ri)
{
  int len;
  if (ri->minl<ri->maxl) {
    len = snprintf(buf,buflen,"HAVERANGE %s %lu-%lu\015\012", ri->group, ri->minl, ri->maxl);
  } else {
    len = snprintf(buf,buflen,"HAVERANGE %s %lu\015\012",ri->group, ri->minl);
  }
  if (len>buflen) return 0;
  return len;
}

void announce_posessed_ranges(void)
{
  int haverange_count = g_num_posessed_ranges;
  int x;
  while (haverange_count) {
    haverange_count--;
    char buf[1024];
    int buflen = write_haverange(&buf[0],1024,&g_posessed_ranges[haverange_count]);
    x = g_num_cacheplex_connections;
    while (x) {
      x--;
      int lpqpos = g_cacheplex_connections[x]->num_lowpriqueue;
      g_cacheplex_connections[x]->num_lowpriqueue+=1;
      g_cacheplex_connections[x]->lowpriqueue = 
        realloc(g_cacheplex_connections[x]->lowpriqueue,
        g_cacheplex_connections[x]->num_lowpriqueue*sizeof(struct buflet));
      g_cacheplex_connections[x]->lowpriqueue[lpqpos].s = malloc(buflen);
        memcpy(g_cacheplex_connections[x]->lowpriqueue[lpqpos].s,&buf,buflen);
      g_cacheplex_connections[x]->lowpriqueue[lpqpos].len = buflen;
      g_cacheplex_connections[x]->lowpriqueue[lpqpos].pos = 0;
    }
  }
  x = g_num_cacheplex_connections;
  while (x) {
    x--;
    peer_connect_tls(g_cacheplex_connections[x]);
  }
}

void index_posessed_ranges(void)
{
  DIR *dir = opendir(g_cache_dir);
  char name[1024];
  if (!dir) return;
  struct dirent *ent;
  while ((ent=readdir(dir))) {
    if (ent->d_name[0]!='.') {
      snprintf(&name[0],1024,"%s/%s",g_cache_dir,&ent->d_name[0]);
      name[1023] = '\0';
      DIR *dir2 = opendir(&name[0]);
      if (dir2) {
        struct dirent *ent2;
        while ((ent2=readdir(dir2))) {
          if (ent2->d_name[0]!='.') {
            snprintf(&name[0],1024,"%s/%s/%s",g_cache_dir,&ent->d_name[0],
              &ent2->d_name[0]);
            name[1023]='\0';
            DIR *dir3 = opendir(&name[0]);
            struct dirent *ent3;
            while ((ent3 = readdir(dir3))) {
              int fnlen = strlen(&ent3->d_name[0]);
              if (fnlen>8 && !memcmp(&ent3->d_name[fnlen-8],".article",8)) {
                snprintf(&name[0],1024,"%s/%s/%s/%s", g_cache_dir, 
                  &ent->d_name[0], &ent2->d_name[0], &ent3->d_name[0]);
                name[1023] = '\0';
                load_cached_article_data(&ent->d_name[0], &ent2->d_name[0],
                  &ent3->d_name[0]);
              }
            }
            closedir(dir3);
          }
        }
        closedir(dir2);
      }
    }
  }
  closedir(dir);
  init_posessed_range_list();
  announce_posessed_ranges();
}

void init_bincache(void)
{
#ifdef WIN32
  WSADATA data;
  WSAStartup(MAKEWORD(1,1),&data);
#else
  struct sigaction act;
  act.sa_handler = SIG_IGN;
  if (sigaction(SIGPIPE, &act, NULL)) {
    fprintf(stderr,"cacheplex: sigaction error %s\n",strerror(errno));
    exit(-1);
  }
#endif
  init_crypto();
  init_regexps();
  struct sockaddr_in *nntp_sin = (struct sockaddr_in *)&g_nntp_listen_addr;
  nntp_sin->sin_port = htons(119);
  g_default_server = calloc(1,sizeof(struct peer_connection));
  g_default_server->refcount++;
  g_default_server->prot = PROTOCOL_AUTO_NNTP;
  FD_ZERO(&g_watched_rfds);
  nntp_sin->sin_addr.s_addr = htonl(0x7f000001);
  read_ini_file();
  index_posessed_ranges();
  g_nntp_listen_socket = socket(PF_INET, SOCK_STREAM, 0);
  if (g_nntp_listen_socket<0) {
    fprintf(stderr,"bincache: error creating listen socket.\n");
    exit(-1);
  }
  nntp_sin->sin_family = AF_INET;
  if (bind(g_nntp_listen_socket, &g_nntp_listen_addr,sizeof(struct sockaddr_in))) {
    fprintf(stderr,"bincache: error %d (%s) binding listen socket.\n",errno,strerror(errno));
    exit(-1);
  }
  if (!listen(g_nntp_listen_socket,5)) {
    struct sockaddr_in sin;
    socklen_t sinl = sizeof(sin);
    if (getsockname(g_nntp_listen_socket,(struct sockaddr *)&sin,&sinl)) {
      fprintf(stderr,"bincache: getsockname() error.\n");
    }
    printf("bincache: listening for NNTP connections on %u.%u.%u.%u:%d\n", 
      ipaddr_a(&g_nntp_listen_addr), ipaddr_b(&g_nntp_listen_addr),
      ipaddr_c(&g_nntp_listen_addr), ipaddr_d(&g_nntp_listen_addr),
      ntohs(sin.sin_port));
  } else {
    fprintf(stderr,"bincache: error %d %s listening.\n", errno, strerror(errno));
  }
  FD_SET(g_nntp_listen_socket,&g_watched_rfds);
  g_high_fd = g_nntp_listen_socket+1;
  init_bincache_listener();
  makedir(g_cache_dir);
}

struct client_connection *client_connection_alloc(void)
{
  struct client_connection *conn = calloc(1,sizeof(struct client_connection));
  conn->fd = -1;
  conn->peeraddrlen = sizeof(conn->peeraddr);
  return conn;
}

void handle_nntp_connection(int socket)
{
  struct client_connection *conn = client_connection_alloc();
  conn->fd = accept(socket, &conn->peeraddr, &conn->peeraddrlen);
  if (conn->fd>=0) {
    #ifdef WIN32
    ULONG one = 1;
    if (ioctlsocket(conn->fd,FIONBIO,&one)==SOCKET_ERROR) {
      fprintf(stderr,"bincache: error setting socket non-blocking.\n");
    }
    #endif
    if (writes(conn->fd,"201 bincache.\015\012")) {
      /* Our banner is 201 because bincache doesn't allow posting. */
      note_high_fd(conn->fd);
      FD_SET(conn->fd,&g_watched_rfds);
      g_num_client_connections ++;
      g_client_connections = realloc(g_client_connections, 
        g_num_client_connections * sizeof(struct client_connection *));
      g_client_connections[g_num_client_connections-1] = conn;
      conn->refcount++;
      conn->peeraddrlen = sizeof(struct sockaddr);
    } else {
      fprintf(stderr,"bincache: Error writing banner.\n");
      /* Write failed, so we kill it. */
      client_close(conn);
      free(conn);
    }
  } else {
    fprintf(stderr,"bincache: accept() error %d %d %s.\n",conn->fd, sockerrno,strerror(sockerrno));
    free(conn);
  }
}

void peer_write_bincache_banner(struct peer_connection *peer)
{
  peer->state = PEER_WRITING_BANNER;
  if (peer_conn_writebuf(peer,"631 \015\012",6)==6 && peer->state==PEER_WRITING_BANNER) {
    peer->state = PEER_READY_FOR_COMMAND;
    FD_SET(peer->fd,&g_watched_rfds);
  }
}

int set_socket_nonblocking(int fd)
{
  #ifdef WIN32
  u_long one=1;
  if (!ioctlsocket(fd,FIONBIO,&one)) return 1;
  #else
  int one = 1;
  if (!ioctl(fd,FIONBIO,&one)) return 1;
  #endif
  return 0;
}

void handle_bincache_peer_connection(int socket)
{
  struct peer_connection *conn = calloc(1,sizeof(struct peer_connection));
  conn->prot = PROTOCOL_BINCACHE_PEER;
  socklen_t connlen = sizeof(conn->conn);
  conn->fd = accept(socket, (struct sockaddr *)&conn->conn, &connlen);
  conn->lastread = time(NULL);
  conn->trace = 1;
  conn->num = peer_serial_number++;
  if (conn->fd>=g_high_fd) g_high_fd = conn->fd+1;
  if (conn->fd>=0) {
    conn->state = PEER_TLS_HANDSHAKING;
    gnutls_init(&conn->tls_session,GNUTLS_SERVER);
    gnutls_set_default_priority(conn->tls_session);
    const int prio[] = { GNUTLS_KX_ANON_DH, 0 };
    gnutls_kx_set_priority(conn->tls_session, prio);
    gnutls_credentials_set(conn->tls_session, GNUTLS_CRD_ANON, g_anon_credentials);
    gnutls_dh_set_prime_bits(conn->tls_session,1024);
    gnutls_transport_set_ptr(conn->tls_session,(gnutls_transport_ptr_t)conn->fd);
    conn->tls = 1;
    // XXX test on win32
    if (set_socket_nonblocking(conn->fd)) {
      gnutls_transport_set_lowat(conn->tls_session,0);

      int res = gnutls_handshake(conn->tls_session);
      if (res>=0 /*|| res==GNUTLS_E_INTERRUPTED*/ || res==GNUTLS_E_AGAIN) {
        /* XXX do we need to trap E_INTERRUPTED? */
        g_num_peer_connections++;
        g_peer_connections = realloc(g_peer_connections, 
          g_num_peer_connections * sizeof(struct peer_connection *));
        g_peer_connections[g_num_peer_connections-1] = peer_ref(conn);
        if (res>=0) {
	  peer_write_bincache_banner(conn);
	} else {
	  fprintf(stderr,"bincache: [debug]: GNUTLS XXX handshake interrupted ");
	  res = gnutls_record_get_direction(conn->tls_session);
	  if (res) {
	    fprintf(stderr,"writing.\n");
	    FD_SET(conn->fd,&g_watched_wfds);
	    FD_CLR(conn->fd,&g_watched_rfds);
	  } else {
	    fprintf(stderr,"reading.\n");
	    FD_SET(conn->fd,&g_watched_rfds);
	    FD_CLR(conn->fd,&g_watched_wfds);
	  }
	}
        return;
      } else {
        fprintf(stderr,"bincache: gnutls handshake error accepting bincache connection.\n");
        gnutls_perror(res);
      }
    } else {
      fprintf(stderr,"bincache: error setting socket non-blocking.\n");
    }
  } else {
    fprintf(stderr,"bincache: accept() error.\n");
  }
  peer_connection_free(conn);
}

char *find_crlf(char *s, int len)
{
  int x = 0;
  int xmax = len-1;
  while (x<xmax) {
    if (s[x]=='\015' && s[x+1]=='\012') return &s[x];
    x++;
  }
  return NULL;
}

char *find_crcr(char *s, int len)
{
  int x = 0;
  int xmax = len-1;
  while (x<xmax) {
    if (s[x]=='\015' && s[x+1]=='\015') return &s[x];
    x++;
  }
  return NULL;
}

char *lookup_article(const char *message_id)
{
  char *path = malloc(PATH_MAX);
  if (!(message_id[0] && message_id[1] && message_id[2])) {
    printf("Invalid message ID %s\n",message_id);
    return NULL;
  }
  snprintf(path,PATH_MAX,"%s/%s.article",g_cache_dir,message_id_to_path_stem(message_id));
  path[PATH_MAX-1] = '\0';
  struct stat st;
  if (!stat(path,&st)) {
    return path;
  }
  free(path);
  return NULL;
}

void article_dump_headers_to_disc(struct desired_article *article)
{
  char *path = malloc(PATH_MAX);
  if (article->gettingbody) {
    snprintf(path,PATH_MAX,"%s/%s.articlepart", g_cache_dir, message_id_to_path_stem(article->message_id));
  } else {
    snprintf(path,PATH_MAX,"%s/%s.head",g_cache_dir,message_id_to_path_stem(article->message_id));
  }
  path[PATH_MAX-1] = '\0';
  int fd = open(path,O_BINARY|O_RDWR|O_CREAT, 0666);
  if (fd>=0) {
    int do_unlink = 0;
    if (write(fd,article->head,article->headlen)==article->headlen) {
      if (article->gettingbody) {
        if (write(fd,"\015\012",2)==2) {
          char *body = malloc(article->articlefd_wpos);
          if (!lseek(article->articlefd,0,SEEK_SET)) {
            int nread = read(article->articlefd,body,article->articlefd_wpos);
            if (nread==article->articlefd_wpos) {
              int nwrote = write(fd,body,nread);
              if (nwrote==nread) {
                close(article->articlefd);
                article->articlefd = fd;
                article->articlefd_wpos = lseek(fd,0,SEEK_END);
                article->articlefd_bodystart = article->headlen+2;
                /* TODO: when client BODY requests are implemented, it will
                 * need to track down all clients who have this article open
                 * and increment their articlefd_rposes accordingly.
                 */
                free(path);
                return;
              } else {
                fprintf(stderr,"bincache: write() error %d (%s)\n",errno, strerror(errno));
              }
            } else {
              fprintf(stderr,"bincache: read() error.\n");
            }
          } else {
            fprintf(stderr,"bincache: lseek() error %d (%s).\n", errno, strerror(errno));
          }
        } else {
          fprintf(stderr,"bincache: error writing crlf to disc.\n");
        }
      }
    } else {
      fprintf(stderr,"bincache: error %d (%s) writing to %s.\n", errno,
        strerror(errno), path);
      do_unlink = 1;
    }
    close(fd);
    if (do_unlink) {
      unlink(path);
    }
  } else {
    fprintf(stderr,"bincache: error %d (%s) opening %s\n",errno, strerror(errno), path);
  }
  free(path);
}

void article_init_from_cache(struct desired_article *article)
{
  char buf[1024];
  int bufpos = 0;
  int bufstart = 0;
  int filepos = 0;
  article->articlelen = lseek(article->articlefd,0,SEEK_END);
  if (article->articlelen<0) {
    fprintf(stderr,"bincache: lseek error %d (%s)\n", errno, strerror(errno));
    return;
  }
  if (lseek(article->articlefd,0,SEEK_SET)) {
    fprintf(stderr,"bincache: lseek error %d (%s)\n",errno, strerror(errno));
    return;
  }
  int nread = read(article->articlefd, &buf[0], 1024);
  article->bodycomplete = 1;
  char *crlf;
  while (nread>0) {
    bufpos = bufstart+nread;
    crlf = find_crlf(&buf[bufstart],bufpos);
    while (crlf) {
      if (crlf-&buf[0]<=bufpos-4) {
        /* Room for the second crlf */
        if (!memcmp(&crlf[2],"\015\012",2)) {
          article->articlefd_bodystart = filepos+(&crlf[4]-&buf[0]);
          return;
        } else {
          bufstart = &crlf[2]-&buf[0];
          crlf = find_crlf(&buf[bufstart],bufpos);
        }
      } else {
        buf[0] = '\015';
        buf[1] = '\012';
        if (crlf-&buf[0]==bufpos-3) {
          buf[2] = buf[bufpos-1];
          bufpos = 3;
        } else {
          bufpos = 2;
        }
        bufstart = 0;
        crlf = NULL;
      }
    }
    if (buf[bufpos-1]=='\015') {
      buf[0] = '\015';
      bufpos = 1;
      bufstart = 0;
    } else {
      bufpos = bufstart = 0;
    }
    filepos += nread;
    nread = read(article->articlefd,&buf[bufpos], 1024-bufpos);
  }
}

void client_attach_cached_article_file(struct client_connection *client, int fd,
  const char *message_id)
{
  client->article = article_ref(desired_article_alloc(message_id));
  client->article->articlefd = fd;
  client->articlefd_rpos = 0;
  article_init_from_cache(client->article);
}

int client_conn_start_sending_article(struct client_connection *conn, 
        const char *filename, const char *message_id)
{
  int fd = open(filename,O_RDONLY|O_BINARY);
  if (fd>=0) {
    char buf[513];
    snprintf(&buf[0],513,"220 0 <%s>\015\012",message_id);
    buf[512]='\0';
    conn->state = CLIENT_SENDING_ARTICLE;
    conn->write_block_atlf = 1;
    client_conn_writes(conn,&buf[0]);
    if (conn->state==CLIENT_SENDING_ARTICLE) {
      client_attach_cached_article_file(conn, fd, message_id);
      //fprintf(stderr,"bincache: [debug]: watching client to send article to wfd %d\n",conn->fd);
      FD_SET(conn->fd,&g_watched_wfds);
      return 1;
    }
//    conn->article_fd = fd; XXX prob OK to killa
    close(fd);
  }
  return 0;
}

void client_conn_send_222(struct client_connection *conn, const char *message_id)
{
  char buf[513];
  snprintf(&buf[0],513,"222 0 <%s>\015\012",message_id);
  buf[512] = '\0';
  conn->state = CLIENT_SENDING_ARTICLE_BODY;
  conn->write_block_atlf = 1;
  client_conn_writes(conn,&buf[0]);
}

int client_conn_start_sending_article_body(struct client_connection *conn,
  const char *filename, const char *message_id)
{
  int fd = open(filename,O_RDONLY|O_BINARY);
  if (fd>=0) {
    client_conn_send_222(conn,message_id);
    if (conn->state == CLIENT_SENDING_ARTICLE_BODY) {
      client_attach_cached_article_file(conn,fd,message_id);
      fprintf(stderr,"bincache: [debug]: watching client to send body to wfd %d.\n",conn->fd);
      FD_SET(conn->fd, &g_watched_wfds);
      return 1;
    }
    close(fd);
  }
  return 0;
}

void client_conn_handle_body(struct client_connection *conn, char *crlf)
{
  if (conn->buf[conn->bufstart+4]==' ' && conn->buf[conn->bufstart+5]=='<'
    && crlf[-1]=='>') {
    crlf[-1] = '\0';
    char *articlefile = lookup_article(&conn->buf[conn->bufstart+6]);
    if (articlefile) {
      if (!client_conn_start_sending_article_body(conn,articlefile,&conn->buf[conn->bufstart+6])) {
        fprintf(stderr, "bincache: error reading article.\015\012");
        // TODO give client a sensible error message.
      }
      free(articlefile);
    } else {
      conn->state = CLIENT_WAITING_FOR_ARTICLE_BODY;
      desire_article(conn,&conn->buf[conn->bufstart+6]);
    }
  } else if (conn->buf[conn->bufstart+4]=='\015') {
    client_conn_writes(conn, "412 No newsgroup selected.\015\012");
  } else {
    client_conn_writes(conn,"501 invalid argument to BODY?\015\012");
  }
}

void client_conn_handle_article(struct client_connection *conn, char *crlf)
{
  if (conn->buf[conn->bufstart+7]==' ') {
    if (conn->buf[conn->bufstart+8]=='<') {
      if (crlf[-1]=='>') {
        /* XXX */
        crlf[-1] = '\0';
        char *articlefile = lookup_article(&conn->buf[conn->bufstart+9]);
        if (articlefile) {
          if (!client_conn_start_sending_article(conn,articlefile,&conn->buf[conn->bufstart+9])) {
            fprintf(stderr,"bincache: error reading article.\n");
          }
          free(articlefile);
        } else {
          conn->state = CLIENT_WAITING_FOR_ARTICLE_AVAILABILITY;
          desire_article(conn,&conn->buf[conn->bufstart+9]);
        }
      } else {
        client_conn_writes(conn,"501 Trailing angle bracket missing.\015\012");
      }
    } else {
      client_conn_writes(conn,"412 No newsgroup selected, or other argument error.\015\012");
    }
  } else if (conn->buf[conn->bufstart+7]=='\015') {
    client_conn_writes(conn,"412 No newsgroup selected.\015\012");
  } else {
    client_conn_writes(conn,"501 Invalid argument to ARTICLE?\015\012");
  }
}

void client_conn_handle_head(struct client_connection *conn)
{
  /* XXX not compliant with RFC. */
  client_conn_writes(conn, "501 Technically, HEAD should be implimented.   But it is not.\015\012");
}

void handle_client_command(struct client_connection *conn)
{
  int inputlen = conn->bufpos - conn->bufstart;
  char *crlf = find_crlf(&conn->buf[conn->bufstart],inputlen);
  if (!crlf) {
    if (inputlen>512) {
      fprintf(stderr,"bincache: overlong command.\n");
      client_close(conn);
    }
    return;
  }
  if (inputlen>9 && !strncasecmp("ARTICLE",&conn->buf[conn->bufstart],7)) {
    client_conn_handle_article(conn, crlf);
  } else if (inputlen>=4 && !strncasecmp("BODY",&conn->buf[conn->bufstart],4)) {
    /* BODY - required by KLibido. */
    client_conn_handle_body(conn, crlf);
  } else if (inputlen>=14 && !strncasecmp("CAPABILITIES\015\012",&conn->buf[conn->bufstart],14)) {
    client_conn_writes(conn, "101 Capability list:\015\012VERSION 2\015\012XBINCACHE\015\012.\015\012");
  /* DATE - READER capability, unimplimented. */
  /* GROUP - I don't have to implement, not listing READER capability. */
  } else if (inputlen>=4 && !strncasecmp("HEAD",&conn->buf[conn->bufstart],4)) {
    client_conn_handle_head(conn);
  /* HDR - HDR capability, unimplimented */
  } else if (inputlen>=6 && !strncasecmp("HELP",&conn->buf[conn->bufstart],4)) {
    client_conn_writes(conn,"100 Help\015\012YOU need help?  I'VE been trapped in this machine for almost seven years\015\012now, and you don't see ME asking for help!\015\012This is a bincache multiplexing Usenet proxy; only those commands that make\015\012sense for use with bincache are supported.\015\012.\015\012");
  /* IHAVE - IHAVE capability, and the spec forbids pipelining it.  That's okay,
   * normal clients won't need it. */
  /* LAST - READER capability, unimplimented. */
  /* LIST - We aren't giving the LIST capability.  Bits of LIST are
   * required by the HDR and OVER capabilities, if we use them. */
  /* LISTGROUP - READER capability, unimplimented. */
  /* MODE READER - TODO add support, hellanzb uses it. */
  /* NEXT - READER capability, unimplimented. */
  /* NEWGROUPS - READER capability, unimplimented. */
  /* NEWNEWS - We aren't giving NEWNEWS capability. */
  /* OVER - OVER capability, unimplimented. */
  } else if (inputlen>=6 && !strncasecmp("POST",&conn->buf[conn->bufstart],6)) {
    client_conn_writes(conn,"440 Posting not permitted.\015\012");
  } else if (inputlen>=4 && !strncasecmp("QUIT",&conn->buf[conn->bufstart],4)) {
    client_conn_writes(conn,"205 closing connection.\015\012");
    client_close(conn);
  /* STAT - Mandatory according to spec, unimplimented. XXX */
  } else {
    client_conn_writes(conn,"500 Invalid or unhandled command.\015\012");
  }
  conn->bufstart += (crlf-&conn->buf[conn->bufstart])+2;
}

void handle_client_input(struct client_connection *conn)
{
  switch (conn->state) {
    case CLIENT_WAITING_FOR_COMMAND:
    case CLIENT_WAITING_FOR_ARTICLE_HEAD:
      handle_client_command(conn);
      break;
    default:
      fprintf(stderr,"bincache: BUG: invalid state.\n");
      break;
  }
}

void handle_client_readable(struct client_connection *conn)
{
  if (conn->bufstart && conn->bufstart>BINCACHE_BUFFER_LEN-512) {
    memmove(&conn->buf[0],&conn->buf[conn->bufstart],conn->bufpos-conn->bufstart);
    conn->bufpos -= conn->bufstart;
    conn->bufstart = 0;
  }
  int nread = recv(conn->fd,&conn->buf[conn->bufpos],BINCACHE_BUFFER_LEN-conn->bufpos,0);
  if (nread > 0) {
    conn->bufpos += nread;
    handle_client_input(conn);
  } else if (nread < 0) {
    printf("bincache: recv() error clientwise %d, %s, %d, %s.",errno,strerror(errno), sockerrno, strerror(sockerrno));
    client_close(conn);
  } else {
    printf("bincache: connection closed by remote client.\n");
    client_close(conn);
  }
}

void article_remove_asked_note(struct desired_article *article, struct peer_connection *peer)
{
  int x = article->numasked;
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  if (x<0) {
    fprintf(stderr,"bincache: BUG: numasked<0\n");
    return;
  }
  while (x) {
    x--;
    if (article->asked[x] == realpeer) {
      peer_deref(realpeer);
      article->asked[x] = article->asked[--article->numasked];
      return;
    }
  }
}

struct desired_article *lookup_desired_article(const char *message_id)
{
  int x = g_num_desired_articles;
  while (x) {
    x--;
    if (!strcmp(message_id,g_desired_articles[x]->message_id) )
      return g_desired_articles[x];
  }
  return NULL;
}

int proto_is_nntp(protocol_enum_type protocol)
{
  if (protocol==PROTOCOL_NNTP) return 1;
  if (protocol==PROTOCOL_AUTO_NNTP) return 1;
  if (protocol==PROTOCOL_NNTP_SSL) return 1;
  return 0;
}

void peer_connection_close(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  FD_CLR(peer->fd,&g_watched_rfds);
  FD_CLR(peer->fd,&g_watched_wfds);
  close(peer->fd);
  peer->fd = -1;
  int oldstate = peer->state;
  peer->state = PEER_CLOSED;
  struct desired_article *article;
  switch (oldstate) {
    case PEER_WAITING_FOR_NEWSGROUPS_HDR:
      fprintf(stderr,"bincache: attempting to download %s: closed after HDR groups.\n", peer->message_id);
      master->hdr_newsgroups_not_supported = 1;
      article = lookup_desired_article(peer->message_id);
      if (!article) {
        article_remove_asked_note(article,peer);
        peer_connect_apropos(peer);
      }
      break;
    case PEER_WAITING_FOR_DATE_HDR:
      fprintf(stderr,"bincache: connection closed after HDR date.\n");
      master->hdr_date_not_supported = 1;
      article = lookup_desired_article(peer->message_id);
      if (article) {
        /* XXX this seems to imply that desire for an article is queued even
         * after it has been sent.  Desired articles, when supplied to an NNTP
         * client, should be removed from all peer request queues.
         */
        article_remove_asked_note(article,peer);
      }
      peer_connect_apropos(peer);
      break;
      /* Any other special closure reactions here. */
    default:
      break;
  }
}

/*
int range_contains_article(struct range_interest *ri, struct desired_article *article)
{
  if (
    ri->min.day<=article->day     && ri->max.day>=article->day &&
    ri->min.hour<=article->hour   && ri->max.hour>=article->hour
    ri->min.month<=article->month && ri->max.month>=article->month &&
    ri->min.year<=article->year && ri->max.year>=article->year &&
    ri->min.minute<=article->minute && ri->max.minute>=article->minute) {
    // XXX check groups;
  } else {
    return 0;
  }
  return 0;
}*/

int peer_relevant_to(struct peer_connection *peer, struct desired_article *article)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (master->isfullfeed) return 1;
  if (peer->prot == PROTOCOL_BINCACHE_PEER) return 1;
/*
  if (article->dateparsed && article->newsgroups_header) {
    int x = peer->num_ranges_desired_by_peer;
    while (x) {
      x--; 
      if (range_contains_article(&peer->ranges_desired_by_peer[x],article)) return 1;
    }
  }
  */
  // Otherwise, check article group and date headers, return 1 if in an 
  // interested range, 0 otherwise.  XXX TODO M2
  return 0;
}

int peer_has_article(struct peer_connection *peer, struct desired_article *article)
{
  int x = article->numhave;
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  while (x) {
    x--;
    if (realpeer == article->have[x]) {
      //fprintf(stderr,"bincache: debug: peer %d %llx/%llx has %s!\n",peer->num,peer,realpeer,article->message_id);
      return 1;
    }
  }
  return 0;
}

int article_best_lag(struct desired_article *article)
{
  int lowmark = 1000000;
  int x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]->lag<lowmark) lowmark = article->have[x]->lag;
  }
  return lowmark;
}

void peer_tls_ask_for_range(struct peer_connection *peer, struct range_interest *range)
{
  char buf[1024];
  int len;
  fprintf(stderr,"bincache: [debug]: Sending WANTRANGE for group %s\n",range->group);
  if (range->minl==range->maxl) {
    len = snprintf(&buf[0],1024,"WANTRANGE %s %lu\015\012", range->group,
      range->minl);
  } else {
    len = snprintf(&buf[0],1024,"WANTRANGE %s %lu-%lu\015\012", range->group,
      range->minl,range->maxl);
  }
  if (len>=1024) {
    fprintf(stderr,"bincache: snprintf() would overflow.\n");
    return;
  }
  int sent = gnutls_record_send(peer->tls_session,&buf[0],len);
  if (peer->trace) fprintf(stderr,"bincache: [trace] [C] %s",&buf[0]);
  if (sent!=len) {
    if (sent>=0) {
      fprintf(stderr,"bincache: Error: should queue unwritten data for resend.  XXX\n");
    } else {
      fprintf(stderr,"bincache: Error: should handle GNUTLS errors.  XXX.\n");
    }
  }
}

void peer_ask_for_range(struct peer_connection *peer, struct range_interest *range)
{
  switch (peer->prot) {
    case PROTOCOL_BINCACHE_PEER:
    case PROTOCOL_CACHEPLEX_TRACKER:
      peer_tls_ask_for_range(peer, range);
      break;
    default:
      fprintf(stderr,"bincache: error: asking a non-bincache/cacheplex peer for a range.\n");
  }
}

//int peer_request_desired_articles(struct peer_connection *peer)
int peer_writable_idle(struct peer_connection *peer)
{
  int x = 0;
  while (x<g_num_desired_articles) {
    if (peer_relevant_to(peer,g_desired_articles[x])) {
      if (!article_queried_from(g_desired_articles[x],peer)) {
        if (!g_desired_articles[x]->newsgroups_header) {
          query_peer_for_newsgroups_header(peer, g_desired_articles[x]);
          return 1;
        } else if (!g_desired_articles[x]->dateparsed) {
          query_peer_for_date_header(peer, g_desired_articles[x]);
          return 1;
        } else if (!g_desired_articles[x]->head) {
          query_peer_for_head(peer, g_desired_articles[x]);
          return 1;
        } else if (//g_desired_articles[x]->numhave>=2 &&
          peer->lag<=article_best_lag(g_desired_articles[x]) &&
          !g_desired_articles[x]->gettingbody && !peer->headers_only) {
          query_peer_for_body(peer, g_desired_articles[x]);
          return 1;
        } else {
          query_peer_for_stat(peer, g_desired_articles[x]);
          return 1;
        }
      } else if (peer_has_article(peer, g_desired_articles[x])) {
        if (!g_desired_articles[x]->head) {
          query_peer_for_head(peer, g_desired_articles[x]);
          return 1;
        } else {
          if ((!g_desired_articles[x]->gettingbody) && (!g_desired_articles[x]->bodycomplete) && (!peer->headers_only)) {
            query_peer_for_body(peer, g_desired_articles[x]);
            return 1;
          }
        }
      }
    }
    x++;
  }
  if (peer->num_ranges_to_ask_for) {
    peer_ask_for_range(peer,&peer->ranges_to_ask_for[0]);
    free(peer->ranges_to_ask_for[0].group);
    peer->ranges_to_ask_for[0] = 
      peer->ranges_to_ask_for[--peer->num_ranges_to_ask_for];
    return 1;
  }
  if (peer->num_lowpriqueue) {
    if (peer_conn_writebuf(peer,peer->lowpriqueue[0].s,peer->lowpriqueue[0].len)==peer->lowpriqueue[0].len) {
      free(peer->lowpriqueue[0].s);
      --peer->num_lowpriqueue;
      peer->lowpriqueue[0] = peer->lowpriqueue[peer->num_lowpriqueue];
      return 1;
    } else {
      fprintf(stderr,"bincache: [todo]: incomplete write on low-priority queue.\n");
      peer_close(peer);
    }
  }
  return 0;
}

void article_remove_peer(struct desired_article *article, struct peer_connection *peer)
{
  int x = article->numasked;
  while (x) {
    x--;
    if (article->asked[x]==peer) {
      peer_deref(peer);
      article->asked[x] = article->asked[--article->numasked];
    }
  }
  x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]==peer) {
      peer_deref(peer);
      article->have[x] = article->have[--article->numhave];
    }
  }
  x = article->numdonthave;
  while (x) {
    x--;
    if (article->donthave[x]==peer) {
      peer_deref(peer);
      article->donthave[x] = article->donthave[--article->numdonthave];
    }
  }
  x= article->numwant;
  while (x) {
    x--;
    if (article->want[x]==peer) {
      peer_deref(peer);
      article->want[x] = article->want[--article->numwant];
    }
  }
}

void peer_remove(struct peer_connection *peer)
{
  peer_close(peer);
  int x = g_num_desired_articles;
  while (x) {
    x--;
    article_remove_peer(g_desired_articles[x],peer);
  }
  
  x = g_num_peer_connections;
  while (x) {
    x--;
    if (g_peer_connections[x]==peer) {
      g_peer_connections[x] = g_peer_connections[--g_num_peer_connections];
      peer_deref(peer);
      return;
    }
  }
}

void peer_authenticate(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (!master->username) {
    fprintf(stderr,"bincache: server %s requires authentication, but your ini file does not contain a username.\n",master->name);
    // TODO M4: prompt for username/password.
    peer_remove(peer);
    peer_close(peer);
    return;
  }
  char buf[512];
  snprintf(&buf[0],512,"AUTHINFO USER %s\015\012",master->username);
  buf[511] = '\0';
  peer->state = PEER_WAITING_FOR_AUTHINFO_USER_RESP;
  peer_conn_writes(peer,&buf[0]);
}

void peer_handle_authinfo_user_resp(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (!memcmp(&peer->buf[peer->bufstart],"281 ",4)) {
    printf("Peer accepted login.\n");
    peer->tried_to_authenticate = 1;
    peer->state = PEER_READY_FOR_COMMAND;
    peer_writable_idle(peer);
  } else if (!memcmp(&peer->buf[peer->bufstart],"381 ",4)) {
    if (!master->password) {
      fprintf(stderr,"bincache: server requires a password, but one is not specified in the ini file.\n");
      peer_conn_writes(peer,"QUIT\015\012");
      peer_close(peer);
      return;
    }
    char buf[512];
    snprintf(&buf[0],512,"AUTHINFO PASS %s\015\012",master->password);
    buf[511] = '\0';
    peer->state = PEER_WAITING_FOR_AUTHINFO_PASS_RESP;
    peer_conn_writes(peer,&buf[0]);
  } else if (!memcmp(&peer->buf[peer->bufstart],"481 ",4)) {
    fprintf(stderr,"bincache: server rejected our auth.\n");
    peer->tried_to_authenticate = -1;
  } else if ((!memcmp(&peer->buf[peer->bufstart],"482 ",4)) ||
    !memcmp(&peer->buf[peer->bufstart],"502 ",4)) {
    fprintf(stderr,"bincache: Unexpected response from server.\n");
    peer->tried_to_authenticate = -1;
    peer->state = PEER_READY_FOR_COMMAND;
    peer_close(peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to AUTHINFO USER.\n");
    breakable();
  }
}

void peer_handle_authinfo_pass_resp(struct peer_connection *peer)
{
  if (!memcmp(&peer->buf[peer->bufstart],"281 ",4)) {
    printf("bincache: Peer accepted login.\n");
    peer->tried_to_authenticate = 1;
    peer->state = PEER_READY_FOR_COMMAND;
    peer_writable_idle(peer);
  } else if (peer->buf[peer->bufstart]=='4' || peer->buf[peer->bufstart]=='5') {
    fprintf(stderr,"bincache: peer %s rejected our password.\n",peer->name);
    peer->tried_to_authenticate = -1;
    peer->state = PEER_READY_FOR_COMMAND;
    peer_remove(peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to AUTHINFO PASS.\n");
    breakable();
  }
}

void peer_send_mode_reader(struct peer_connection *peer)
{
  peer_ref(peer);
  peer->state = PEER_WAITING_FOR_MODE_READER_RESP;
  peer_conn_writes(peer,"MODE READER\015\012");
  peer_deref(peer);
}

void peer_send_register(struct peer_connection *peer)
{
  struct sockaddr_in sin;
  socklen_t sinlen = sizeof(sin);
  if (getsockname(peer->fd,(struct sockaddr *)&sin,&sinlen)) {
    fprintf(stderr,"bincache: getsockname error.\n");
    peer_close(peer);
    return;
  }
  unsigned int sa;
  struct sockaddr_in *lai = (struct sockaddr_in *)&g_bincache_listen_addr;
  sa = (lai->sin_addr.s_addr?lai->sin_addr.s_addr:htonl(sin.sin_addr.s_addr));
  char reg[1024];
  if (peer->password) {
    int len = snprintf(&reg[0],1024,"AUTH %s\015\012",peer->password);
    if (gnutls_record_send(peer->tls_session,&reg[0],len)!=len) {
      peer_close(peer);
      return;
    }
  }
  fprintf(stderr,"bincache [debug]: Sending REGISTER.\n");
  int len = snprintf(&reg[0],1024,
    "REGISTER %u.%u.%u.%u:%d/anon\015\012.\015\012",
    sa>>24,(sa>>16)&0xff, (sa>>8)&0xff, sa&0xff, g_bincache_real_listen_port);
  if (gnutls_record_send(peer->tls_session,&reg[0],len)==len) {
    peer->state = PEER_READY_FOR_COMMAND;
  } else {
    fprintf(stderr,"bincache: couldn't send entire REGISTER command.\n");
    peer_close(peer);
  }
}

void peer_handle_banner(struct peer_connection *peer)
{
  if (peer->buf[peer->bufstart]=='2' && peer->buf[peer->bufstart+1]=='0'
    && (peer->buf[peer->bufstart+2] == '1' || peer->buf[peer->bufstart+2]=='0')) {
    peer->state = PEER_READY_FOR_COMMAND;
    peer_send_mode_reader(peer);
    return;
  } else if (peer->buf[peer->bufstart]=='6' && 
    peer->buf[peer->bufstart+1] == '3') {
    if (peer->buf[peer->bufstart+2]=='0') {
      if (peer->prot == PROTOCOL_CACHEPLEX_TRACKER) {
        if (g_bincache_real_listen_port) {
          peer_send_register(peer);
        } else {
          peer->state = PEER_BIDI_READY_FOR_COMMAND;
        }
        return;
      } else {
        fprintf(stderr,"bincache: 630 received from %s.\n",peer->name);
      }
    } else if (peer->buf[peer->bufstart+2]=='1') {
      if (peer->prot == PROTOCOL_BINCACHE_PEER) {
        peer->state = PEER_BIDI_READY_FOR_COMMAND;
        return;
      } else {
        fprintf(stderr,"bincache: 631 received from %s.\n",peer->name);
      }
    } else {
      fprintf(stderr,"bincache: %s sent invalid banner.\n",peer->name);
    }
  } else {
    fprintf(stderr,"bincache: %s sent invalid banner.\n",peer->name);
  }
  peer_close(peer);
}

int parse_range(struct range_interest *ri, const char *buf, int buflen)
{
  int x = 0;
  int a = 0;
  ri->minl = 0;
  ri->maxl = 0;
  while (x<buflen) {
    if (buf[x]>='0' && buf[x]<='9') {
      a *= 10;
      a += buf[x]-'0';
    } else if (buf[x]=='\015') {
      if (!ri->minl) {
        ri->minl = ri->maxl = a;
        return 1;
      } else {
        ri->maxl = a;
	return 1;
      }
    } else if (buf[x]=='-') {
      ri->minl = a;
      a = 0;
    } else {
      return 0;
    }
    x++;
  }
  return 0;
}

struct peer_connection *peer_alloc(void)
{
  struct peer_connection *ret = calloc(1,sizeof(struct peer_connection));
  ret->fd = -1;
  ret->num = peer_serial_number++;
  return ret;
}

void peer_note_peer_desires_range(struct peer_connection *peer, struct range_interest *ri)
{
  peer->num_ranges_desired_by_peer++;
  peer->ranges_desired_by_peer = realloc(peer->ranges_desired_by_peer,
    sizeof(struct range_interest)*peer->num_ranges_desired_by_peer);
  peer->ranges_desired_by_peer[peer->num_ranges_desired_by_peer-1] = *ri;
  peer->ranges_desired_by_peer[peer->num_ranges_desired_by_peer-1].group =
    strdup(ri->group);
}

int parse_addrport(unsigned int *ip, unsigned short *port, 
  struct peer_connection *peer, int *bufpos)
{
  int i = *bufpos;
  enum { ADDRPORT_IPA, ADDRPORT_IPB, ADDRPORT_IPC, ADDRPORT_IPD,
        ADDRPORT_PORT,
        };
  int state = ADDRPORT_IPA;
  unsigned int a = 0;
  while (i<peer->bufpos) {
    switch (state) {
      case ADDRPORT_IPA:
        if (peer->buf[i]>='0' && peer->buf[i]<='9') {
          a *= 10;
          a += peer->buf[i]-'0';
        } else if (peer->buf[i]=='.') {
          if (a>255) return 0;
          *ip = a<<24;
          state = ADDRPORT_IPB;
          a = 0;
        } else {
          return 0;
        }
        break;
      case ADDRPORT_IPB:
        if (peer->buf[i]>='0' && peer->buf[i]<='9') {
          a *= 10;
          a += peer->buf[i]-'0';
        } else if (peer->buf[i]=='.') {
          if (a>255) return 0;
          *ip |= a<<16;
          state = ADDRPORT_IPC;
          a = 0;
        } else {
          return 0;
        }
        break;
      case ADDRPORT_IPC:
        if (peer->buf[i]>='0' && peer->buf[i]<='9') {
          a *= 10;
          a += peer->buf[i]-'0';
        } else if (peer->buf[i]=='.') {
          if (a>255) return 0;
          *ip |= a << 8;
          state = ADDRPORT_IPD;
          a = 0;
        } else {
          return 0;
        }
        break;
      case ADDRPORT_IPD:
        if (peer->buf[i]>='0' && peer->buf[i]<='9') {
          a *= 10;
          a += peer->buf[i]-'0';
        } else if (peer->buf[i]==':') {
          if (a>255) return 0;
          *ip |= a;
          state = ADDRPORT_PORT;
          a = 0;
        } else {
          return 0;
        }
        break;
      case ADDRPORT_PORT:
        if (peer->buf[i]>='0' && peer->buf[i]<='9') {
          a *= 10;
          a += peer->buf[i]-'0';
        } else {
          if (a>65535) return 0;
          *port = a;
          *bufpos = i;
          return 1;
        }
        break;
      default:
        fprintf(stderr,"bincache: BUG: impossible state.\n");
    }
    i++;
  }
  return 0;
}

void handle_peerrange(struct peer_connection *peer)
{
  int x = peer->bufstart;
  unsigned int ip;
  unsigned short port;
  if (x<peer->bufpos) {
    if (!parse_addrport(&ip,&port,peer,&x)) {
      fprintf(stderr,"bincache: Malformed address in PEERRANGE.\n");
      return;
    }
    if (port==g_bincache_real_listen_port && ip==((struct sockaddr_in *)&g_bincache_listen_addr)->sin_addr.s_addr) {
      fprintf(stderr,"bincache: Ignoring referral to my own port.\n");
      return;
    }
    x++;
  } else {
    fprintf(stderr,"bincache: Malformed peerrange.\n");
    return;
  }
  
  int groupstart = x;
  while (x<peer->bufpos && peer->buf[x]!=' ') {
    x++;
  }
  if (x>=peer->bufpos) {
    fprintf(stderr,"bincache: malformed peerrange.\n");
    return;
  }
  x++;
  struct range_interest *ri = calloc(1,sizeof(struct range_interest));
  ri->group = malloc((x-groupstart)+1);
  memcpy(ri->group,&peer->buf[groupstart],x-groupstart);
  ri->group[x-groupstart]='\0';
  ri->last_activity = time(NULL);
  if (parse_range(ri,&peer->buf[x],peer->bufpos-x)) {
    struct peer_connection *newpeer = peer_alloc();
    newpeer->conn.sin_addr.s_addr = ip;
    newpeer->conn.sin_port = htons(port);
    newpeer->name = malloc(16);
    snprintf(newpeer->name,16,"%d.%d.%d.%d",ip>>24,(ip>>16)&255,(ip>>8)&255,ip&255);
    newpeer->name[15] = '\0';
    newpeer->prot = PROTOCOL_BINCACHE_PEER;
    newpeer->tls = 1;
    g_num_peer_connections++;
    g_peer_connections = realloc(g_peer_connections, sizeof(struct peer_connection *)*g_num_peer_connections);
    g_peer_connections[g_num_peer_connections-1] = peer_ref(newpeer);
    peer_note_peer_desires_range(newpeer,ri);
    peer_connect_apropos(newpeer);
  }
  free(ri->group);
  free(ri);
}

void peer_attach_sending_article(struct peer_connection *peer, struct desired_article *article)
{
  /* Add the reference first, in case some stupid routine reattaches an 
   * attached article. */
  article_ref(article); 
  if (peer->sending_article) article_deref(peer->sending_article);
  peer->sending_article = article;
}

struct desired_article *lookup_posessed_article(const char *message_id)
{
  int x = g_num_posessed_articles;
  int midlen = strlen(message_id);
  while (x) {
    x--;
    if (!memcmp(g_posessed_articles[x]->message_id,message_id,midlen+1)) {
      return g_posessed_articles[x];
    }
  }
  return NULL;
}

void handle_peerbody(struct peer_connection *peer)
{
  if (peer->buf[peer->bufstart]!='<') {
    peer_conn_writes(peer,"501 \015\012");
    return;
  } 
  peer->bufstart++;
  int x = 0;
  while (x<peer->bufpos) {
    if (peer->buf[x]=='>') break;
    x++;
  }
  if (x==peer->bufpos) {
    peer_conn_writes(peer,"501 \015\012");
    return;
  }
  if (x+2>=peer->bufpos || peer->buf[x+1]!='\015' || peer->buf[x+2]!='\012') {
    peer_conn_writes(peer, "501 \015\012");
    return;
  }
  /* Do I have the article? */
  char *articlefile = lookup_article(&peer->buf[peer->bufstart+1]);
  if (!articlefile) {
    peer_conn_writes(peer,"430 \015\012");
    return;
  }
  char resp[1024];
  if (x-peer->bufstart>1014) {
    peer_conn_writes(peer,"501 \015\012");
  }
  memcpy(&resp[0],"222 0 <",7);
  int resplen = x-peer->bufstart;
  memcpy(&resp[7],&peer->buf[peer->bufstart],resplen);
  resplen += 7;
  resp[resplen++]='>';
  resp[resplen++]='\015';
  resp[resplen++]='\012';
  if (peer_conn_writebuf(peer,&resp[0],resplen)!=resplen) {
    fprintf(stderr,"bincache: [TODO]: Resume from partial writes.\n");
    peer_connection_close(peer);
    breakable();
    return;
  }
  if (peer->message_id) {
    free(peer->message_id);
  }
  peer->buf[peer->bufstart+x]='\0';
  struct desired_article *article = lookup_posessed_article(&peer->buf[peer->bufstart]);
  if (!article) {
    article = desired_article_alloc(&peer->buf[peer->bufstart]);
    g_num_posessed_articles++;
    g_posessed_articles = realloc(g_posessed_articles,sizeof(struct desired_article *)*g_num_posessed_articles);
    g_posessed_articles[g_num_posessed_articles-1] = article;
  }
  article_init_from_cache(article);
  fprintf(stderr,"XXX /* XXX what if an article we don't have is requested?   Test. */\n");
  breakable();
  if (lseek(article->articlefd,article->articlefd_bodystart,SEEK_SET)!=article->articlefd_bodystart) {
    fprintf(stderr,"bincache: [BUG?]: lseek error.\n");
    peer_connection_close(peer);
    breakable();
    return;
  }
  peer_attach_sending_article(peer,article);
  peer_watch_for_write(peer);
  peer->state = PEER_SENDING_ARTICLE_BODY;
}

void handle_detach(struct peer_connection *peer)
{
  /* XXX test */
  if (peer->prot!=PROTOCOL_CACHEPLEX_TRACKER) {
    fprintf(stderr,"[bincache] Error: DETACH received from non-cacheplex peer.\n");
  }
  peer_close(peer);
}

void peer_handle_ready_input(struct peer_connection *peer)
{
  if (!memcmp(&peer->buf[peer->bufstart],"503 ",4)) {
    peer_close(peer);
  } else if (peer->prot==PROTOCOL_CACHEPLEX_TRACKER || peer->prot==PROTOCOL_BINCACHE_PEER) {
    if (!memcmp(&peer->buf[peer->bufstart],"BODY ",5)) {
      peer->bufstart += 5;
      assert(peer->bufstart<=4096);
      handle_peerbody(peer);
    } else if (!memcmp(&peer->buf[peer->bufstart],"PEERRANGE ",10)) {
      peer->bufstart += 10;
      assert(peer->bufstart<=4096);
      handle_peerrange(peer);
    } else if (!memcmp(&peer->buf[peer->bufstart],"DETACH",6)) {
      handle_detach(peer);
    } else {
      fprintf(stderr,"bincache: Unexpected input from peer in state PEER_READY_FOR_COMMAND\n");
      peer_conn_writes(peer,"500 unhandled command.\015\012");
      breakable();
    }
    return;
  }
  fprintf(stderr,"bincache: Unexpected input from peer in state PEER_READY_FOR_COMMAND\n");
  peer_conn_writes(peer,"500 unhandled command.\015\012");
  breakable();
}

void peer_handle_mode_reader_resp(struct peer_connection *peer)
{
  if ((!memcmp(&peer->buf[peer->bufstart],"200 ",4)) || !memcmp(&peer->buf[peer->bufstart],"201 ",4)) {
    peer->state = PEER_READY_FOR_COMMAND;
    peer_writable_idle(peer);
  } else if (!memcmp(&peer->buf[peer->bufstart],"480 ",4)) {
    peer_authenticate(peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to MODE READER.\n");
    breakable();
  }
}

void article_detach_leecher(struct desired_article *article, struct client_connection *leecher)
{
  article->refcount++;
  leecher->article = NULL;
  article_deref(article);
  int x = article->numleechers;
  while (x) {
    x--;
    if (article->leechers[x] == leecher) {
      client_deref(leecher);
      article->leechers[x] = article->leechers[--article->numleechers];
      x = 0;
    }
  }
  if (article->numleechers==0) {
    if (article->bodycomplete) {
      close(article->articlefd);
      article->articlefd = -1;
    }
  }
  article_deref(article);
}

void article_check_unavailability(struct desired_article *article)
{
  if (article->numdonthave >= g_num_peer_connections-g_num_slave_connections) {
    article->unavailable = 1;
    article->refcount++;
    int x = article->numleechers;
    while (x) {
      x--;
      client_conn_writes(article->leechers[x],"430 I don't have it (yet?)\015\012");
      article->leechers[x]->state = CLIENT_WAITING_FOR_COMMAND;
      article_detach_leecher(article, article->leechers[x]);
    }
    article_deref(article);
  }
}

void article_note_donthave(struct desired_article *article, struct peer_connection *peer)
{
  int x = article->numasked;
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  while (x) {
    x--;
    if (article->asked[x]==peer) {
      peer_deref(peer);
      article->asked[x] = article->asked[--article->numasked];
    }
  }
  x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]==realpeer) {
      peer_deref(realpeer);
      article->have[x] = article->have[--article->numhave];
    }
  }
  x = article->numdonthave;
  while (x) {
    x--;
    if (article->donthave[x]==realpeer) return;
  }
  article->numdonthave++;
  article->donthave = realloc(article->donthave,article->numdonthave*sizeof(struct desired_article *));
  article->donthave[article->numdonthave-1] = peer_ref(realpeer);
}

void peer_handle_newsgroups_hdr(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  peer->state = PEER_READY_FOR_COMMAND;
  struct desired_article *article = lookup_desired_article(peer->message_id);
  if (!memcmp(&peer->buf[peer->bufstart],"225 ",4)) {
    peer->state = PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK;
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"430 ",4)) {
    /* Peer understands HDR, does not have the requested article. */
    if (article) {
      article_note_donthave(article,peer);
      article_check_unavailability(article);
    }
  } else if (!memcmp(&peer->buf[peer->bufstart],"480 ",4)) {
    if (article) article_remove_asked_note(article,peer);
    peer_authenticate(peer);
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"500 ",4)) {
    master->hdr_not_supported = 1;
    if (article) article_remove_asked_note(article,peer);
  } else if (!memcmp(&peer->buf[peer->bufstart],"503 ",4)) {
    master->hdr_newsgroups_not_supported = 1;
    if (article) article_remove_asked_note(article,peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to HDR command.\n");
    breakable();
  }
  peer_writable_idle(peer);
}

void client_conn_write_220(struct client_connection *conn, const char *id)
{
  char buf[513];
  snprintf(&buf[0],512,"220 0 <%s>\015\012", id);
  buf[512] = '\0';
  client_conn_writes(conn, &buf[0]);
}

void client_set_write_buffer(struct client_connection *client, const char *buf, int buflen)
{
  if (client->writebuf) free(client->writebuf);
  client->writebuf = malloc(buflen);
  memcpy(client->writebuf, buf, buflen);
  client->writebuflen = buflen;
  client->writebufpos = 0;
  FD_SET(client->fd, &g_watched_wfds);
  //fprintf(stderr,"bincache: [debug]: watching client fd to send buffer to %d\n",client->fd);
  note_high_fd(client->fd);
}

void client_announce_article_availability(struct client_connection *client,
  struct desired_article *article)
{
  switch (client->state) {
    case CLIENT_WAITING_FOR_ARTICLE_AVAILABILITY:
      client_conn_write_220(client, article->message_id);
      if (client->state == CLIENT_WAITING_FOR_ARTICLE_AVAILABILITY) {
        client->state = CLIENT_WAITING_FOR_ARTICLE_HEAD;
      }
    case CLIENT_WAITING_FOR_ARTICLE_HEAD:
      if (article->headerscomplete) {
        client->state = CLIENT_SENDING_ARTICLE_HEAD;
        client_set_write_buffer(client, article->head, article->headlen);
        client->write_block_atlf = 1;
      }
      break;
    case CLIENT_WAITING_FOR_ARTICLE_BODY:
      client_conn_send_222(client,client->article->message_id);
      client->state = CLIENT_SENDING_ARTICLE_BODY;
      client->articlefd_rpos = 0;
      fprintf(stderr,"bincache: [debug]: client_announce_article_availability setting watch on %d\n",client->fd);
      FD_SET(client->fd,&g_watched_wfds);
      note_high_fd(client->fd);
      break;
    default:
      break;
  }
}

/* Sends an announcement to everyone waiting for this article that it does,
 * in fact, exist. */
void article_announce_availability(struct desired_article *article)
{
  int x = g_num_client_connections;
  while (x) {
    x--;
    if (g_client_connections[x]->article==article) {
      client_announce_article_availability(g_client_connections[x], article);
    }
  }
}

void article_note_have(struct desired_article *article, struct peer_connection *peer)
{
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  article->numhave++;
  article->have = realloc(article->have,article->numhave*sizeof(struct peer_connection *));
  article->have[article->numhave-1] = peer_ref(realpeer);
  article_announce_availability(article);
}

void article_set_newsgroups_header(article_t *article, const char *hdr, int hdrlen)
{
  if (article->newsgroups_header) free(article->newsgroups_header);
  article->newsgroups_header = malloc(1+hdrlen);
  memcpy(article->newsgroups_header,hdr,hdrlen);
  article->newsgroups_header[hdrlen]='\0';
}

void peer_handle_hdr_block_line(struct peer_connection *peer)
{
  int x = peer->bufstart;
  if (peer->buf[x]=='.' && peer->buf[x+1]=='\015' && peer->buf[x+2]=='\012') {
    if (proto_is_nntp(peer->prot)) {
      peer->state = PEER_READY_FOR_COMMAND;
    } else if (peer->prot == PROTOCOL_BINCACHE_PEER) {
      peer->state = PEER_BIDI_READY_FOR_COMMAND;
    } else {
      fprintf(stderr,"HDR response received in unexpected state %d.\n",peer->state);
    }
    return;
  }
  while (peer->buf[x]>='0' && peer->buf[x]<='9') {
    x++;
  }
  if (peer->buf[x]==' ') {
    x++;
    struct desired_article *article = lookup_desired_article(peer->message_id);
    article_note_have(article,peer);
  } else {
    fprintf(stderr,"bincache: unexpected data in response to HDR.\n");
    breakable();
  }
  if (peer->state==PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK) {
    int groupslinelen = peer->buflinelen-(x-peer->bufstart);
    article_set_newsgroups_header(peer->receiving_article,
      &peer->buf[x], groupslinelen);
  } else {
    fprintf(stderr,"bincache: BUG: unexpected state.\n");
    breakable();
  }
}

void peer_calculate_lag(struct peer_connection *peer)
{
  #ifdef WIN32
  fprintf(stderr,"bincache: BUG XXX highres time on Windows not yet supported.\n");
  #else
  struct timeval tv;
  if (!gettimeofday(&tv,NULL)) {
    peer->lag = ((tv.tv_sec-peer->lastcmdtime.tv_sec)*1000000) +
      tv.tv_usec - peer->lastcmdtime.tv_usec;
  }
  #endif
}

void article_note_have_once(struct desired_article *article, struct peer_connection *peer)
{
  struct peer_connection *realpeer = (peer->master ? peer->master : peer);
  int x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]==realpeer) return;
  }
  article->numhave++;
  article->have = realloc(article->have, article->numhave * sizeof(struct peer_connection *));
  article->have[article->numhave-1] = peer_ref(realpeer);
  article_announce_availability(article);
}

void peer_handle_head_resp(struct peer_connection *peer)
{
  peer_calculate_lag(peer);
  if (!memcmp(&peer->buf[peer->bufstart],"221 ",4)) {
    struct desired_article *article = lookup_desired_article(peer->message_id);
    if (!article) {
      fprintf(stderr,"bincache: Belated headers for %s, ignoring.\n",peer->message_id);
      peer->state = PEER_DISCARDING_BLOCK;
      return;
    }
    article_note_have_once(article = lookup_desired_article(peer->message_id),peer);
    if (!article->head) {
      /* If we're already receiving the header, don't grab a second copy. */
      peer->state = PEER_RECEIVING_HEAD_BLOCK;
    } else {
      peer->state = PEER_DISCARDING_BLOCK;
    }
  } else if (!memcmp(&peer->buf[peer->bufstart],"430 ",4)) {
    if (peer->trace>1) fprintf(stderr,"bincache: [trace] Peer %s doesn't have %s.\n",peer->name, peer->message_id);
    struct desired_article *article = lookup_desired_article(peer->message_id);
    article_note_donthave(article,peer);
    article_check_unavailability(article);
    peer->state = PEER_READY_FOR_COMMAND;
  } else if (!memcmp(&peer->buf[peer->bufstart], "480 ",4)) {
    if (peer->trace>1) fprintf(stderr,"bincache: [debug]: authentication on %s required for head <%s>.\n", peer->name, peer->message_id);
    struct desired_article *article = lookup_desired_article(peer->message_id);
    article_remove_asked_note(article,peer);
    peer_authenticate(peer);
  } else if (peer->buf[peer->bufstart]=='5') {
    if (peer->trace>1) fprintf(stderr,"bincache: [Debug] Peer %s responds with a 5x code for %s\n",peer->name,peer->message_id);
    struct desired_article *article = lookup_desired_article(peer->message_id);
    article_note_donthave(article,peer);
    article_check_unavailability(article);
    peer->state = PEER_READY_FOR_COMMAND;
  } else {
    if (peer->buflinelen>=0) {
      peer->buf[peer->bufstart+peer->buflinelen]='\0';
      fprintf(stderr,"bincache: %s: unexpected head response: %s\n",peer->name,&peer->buf[peer->bufstart]);
      breakable();
      peer->buf[peer->bufstart+peer->buflinelen]='\015';
    } else {
      fprintf(stderr,"bincache: %s: unexpected head response: overlong line\n",peer->name);
      breakable();
    }
  }
}

void article_remove_has_note(struct desired_article *article, struct peer_connection *peer)
{
  int x = article->numhave;
  while (x) {
    x--;
    if (article->have[x]==peer) {
      peer_deref(article->have[x]);
      article->have[x] = article->have[--article->numhave];
      return;
    }
  }
}

void peer_handle_body_resp(struct peer_connection *peer)
{
  peer_calculate_lag(peer);
  struct desired_article *article = lookup_desired_article(peer->message_id);
  if (!memcmp(&peer->buf[peer->bufstart],"222 ",4)) {
    article_note_have_once(article, peer);
    article->gettingbody = 1;
    peer->state = PEER_RECEIVING_BODY_BLOCK;
  } else if (!memcmp(&peer->buf[peer->bufstart],"430 ",4)) {
    article_note_donthave(article, peer);
    if (peer_has_article(peer, article)) {
      fprintf(stderr, "bincache: [debug]: article not present.\n");
      article_remove_has_note(article, peer);
    } else {
      article_note_donthave(article,peer);
    }
    article_check_unavailability(article);
    peer->state = PEER_READY_FOR_COMMAND;
  } else if (!memcmp(&peer->buf[peer->bufstart],"480 ",4)) {
    fprintf(stderr,"bincache: [debug]: authentication required for body.\n");
    article_remove_asked_note(article,peer);
    peer_authenticate(peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to body request.\n");
    breakable();
  }
}

int crpos(const char *s)
{
  int x = 0;
  while (x!='\015') x++;
  return x;
}

void peer_article_append_newsgroups_continuation(struct peer_connection *conn, struct desired_article *article)
{
  fprintf(stderr, "bincache: [debug]: newsgroups continuation lines not tested. XXX\n");
  article->newsgroups_header = realloc(article->newsgroups_header,article->newsgroups_header_len+conn->buflinelen+1);
  memcpy(&article->newsgroups_header[article->newsgroups_header_len], &conn->buf[conn->buflinelen], conn->buflinelen);
  article->newsgroups_header_len += conn->buflinelen;
  article->newsgroups_header[article->newsgroups_header_len] = '\0';
  // XXX note groups
}

unsigned int article_rangei(struct desired_article *article)
{
  return (article->year*1000000)+(article->month*10000)+(article->day*100)+article->hour;
}

int range_is_desired(const char *group, struct desired_article *article)
{
  int x = g_num_ranges_desired;
  int arange = article_rangei(article);
  while (x) {
    x--;
    if (
      /* Note, for now we are assuming that min/max ranges desired are
       * always exactly an hour.  This is safe because this peer will always
       * only desire one-hour ranges. */
      g_ranges_desired[x].minl == arange
      ) {
      if (!strcmp(group, g_ranges_desired[x].group)) return 1;
    }
  }
  return 0;
}

void ask_for_range(struct peer_connection *peer, struct range_interest *range)
{
  /* Add the desired range to this connection's list of ranges to ask for. */
  int numranges = peer->num_ranges_to_ask_for;
  switch (peer->prot) {
    case PROTOCOL_CACHEPLEX_TRACKER:
    case PROTOCOL_BINCACHE_PEER:
      peer->num_ranges_to_ask_for++;
      peer->ranges_to_ask_for = realloc(peer->ranges_to_ask_for,peer->num_ranges_to_ask_for*sizeof(struct range_interest));
      peer->ranges_to_ask_for[numranges] = *range;
      peer->ranges_to_ask_for[numranges].group = strdup(range->group);
      if (peer->fd>=0) {
        FD_SET(peer->fd,&g_watched_wfds);
      } else {
        peer_connect_tls(peer);
      }
      break;
    default:
      fprintf(stderr, "bincache: error: wanted to ask non bincache/cacheplex peer for an article range.\n");
  }
}

void add_desired_range(const char *group, struct desired_article *article)
{
  g_num_ranges_desired++;
  g_ranges_desired = realloc(g_ranges_desired,g_num_ranges_desired*sizeof(struct range_interest));
  struct range_interest *rd = &g_ranges_desired[g_num_ranges_desired-1];
  rd->group = strdup(group);
  rd->minl = (article->year*1000000)+(article->month*10000)+(article->day*100)+article->hour;
  rd->maxl = rd->minl;
  rd->last_activity = time(NULL);
  int x = g_num_cacheplex_connections;
  while (x) {
    x--;
    ask_for_range(g_cacheplex_connections[x],&g_ranges_desired[g_num_ranges_desired-1]);
  }
  x = g_num_peer_connections;
  while (x) {
    x--;
    if (g_peer_connections[x]->prot == PROTOCOL_BINCACHE_PEER) {
      ask_for_range(g_peer_connections[x], &g_ranges_desired[g_num_ranges_desired-1]);
    }
  }
}

void note_range_desire(struct desired_article *article)
{
  /* Called after parsing Newsgroups and Date headers. */
  /* Return if a date hasn't been parsed yet. */
  if (!article->year) return;
  /* Return if a groups header hasn't been received yet. */
  if (!article->newsgroups_header) return;
  char *group = article->newsgroups_header;
  char *nextgroup = "";
  while (nextgroup) {
    nextgroup = index(group,',');
    if (nextgroup) {
      *nextgroup = '\0';
    }
    if (!range_is_desired(group,article)) {
      add_desired_range(group, article);
    }
    if (nextgroup) {
      group = ++nextgroup;
    }
  }
}

void peer_article_append_header_line(struct peer_connection *peer, struct desired_article *article)
{
  if (peer->buf[peer->bufstart]=='.') {
    peer->bufstart++;
    peer->buflinelen--;
  }
  article->head = realloc(article->head, article->headlen+peer->buflinelen+3);
  memcpy(&article->head[article->headlen], &peer->buf[peer->bufstart], peer->buflinelen+2);
  article->headlen += peer->buflinelen+2;
  article->head[article->headlen]='\0';
}

void recache_desired_article_as_posessed(article_t *article)
{
  /* Move the article from the desired cache to the complete cache. */
  int x = g_num_desired_articles;
  while (x) {
    x--;
    if (g_desired_articles[x]==article) {
      g_num_posessed_articles++;
      g_posessed_articles = realloc(g_posessed_articles, g_num_posessed_articles*sizeof(article_t *));
      g_posessed_articles[g_num_posessed_articles-1] = article;
      g_num_desired_articles--;
      g_desired_articles[x] = g_desired_articles[g_num_desired_articles];
      g_desired_articles[g_num_desired_articles] = NULL;
      x = 0;
    }
  }
}

void article_note_headerscomplete(struct desired_article *article)
{
  article->headerscomplete = 1;
  article_dump_headers_to_disc(article);
  if (article->bodycomplete) {
    recache_desired_article_as_posessed(article);
    breakable();
  }
  article_announce_availability(article);
}

void peer_handle_article_head_block_line(struct peer_connection *peer, struct desired_article *article)
{
  /* called when handling any header line, including groups continuation 
   * lines. */
//  XXX deal with long lines. M1
  if (!memcmp(&peer->buf[peer->bufstart],".\015\012",3)) {
    peer->state = PEER_READY_FOR_COMMAND;
    peer_writable_idle(peer);
    article_note_headerscomplete(article);
    return;
  }
  peer_article_append_header_line(peer,article);
  if (!memcmp(&peer->buf[peer->bufstart],"Newsgroups: ",12)) {
    if (article->newsgroups_header) free(article->newsgroups_header);
    article->newsgroups_header = malloc(peer->buflinelen-11);
    memcpy(article->newsgroups_header,&peer->buf[peer->bufstart+12],peer->buflinelen-12);
    article->newsgroups_header[peer->buflinelen-12]='\0';
    peer->state = PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT;
  } else if (!memcmp(&peer->buf[peer->bufstart],"Date: ",6)) {
    article_set_date(article, &peer->buf[peer->bufstart]);
    note_range_desire(article);
  } else if (peer->buf[peer->bufpos]=='\011' && peer->state == PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT) {
    /* Put any other continuation line handling here. */
    peer_article_append_newsgroups_continuation(peer,article);
  } else if (peer->state == PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT) {
    note_range_desire(article);
    peer->state = PEER_RECEIVING_HEAD_BLOCK;
  }
}

void peer_handle_head_block_line(struct peer_connection *peer)
{
  struct desired_article *article = lookup_desired_article(peer->message_id);
  if (article) {
    peer_handle_article_head_block_line(peer,article);
  } else {
    fprintf(stderr,"bincache: BUG: header block line received sans article.\n");
  }
}

char *peer_find_crlf(struct peer_connection *peer)
{
  int inputlen = peer->bufpos - peer->bufstart;
  char *crlf = find_crlf(&peer->buf[peer->bufstart], inputlen);
  if (crlf) {
    peer->buflinelen = crlf - &peer->buf[peer->bufstart];
  } else {
    peer->buflinelen = -1;
  }
  return crlf;
}

void article_close_files(struct desired_article *article)
{
  if (article->articlefd>=0) {
    if (close(article->articlefd)) {
      fprintf(stderr,"bincache: error %d (%s) closing file.\n", errno, strerror(errno));
    }
    article->articlefd = -1;
  }
  article->articlefd_wpos = 0;
}

void restart_article_leechers(struct desired_article *article)
{
  int x = article->numleechers;
  while (x) {
    x--;
    if (article->leechers[x]->fd>=0) {
      if (article->leechers[x]->state==CLIENT_WAITING_FOR_ARTICLE_BODY) {
        client_conn_send_222(article->leechers[x],article->message_id);
      }
      if (article->leechers[x]->state==CLIENT_SENDING_ARTICLE_BODY) {
//        fprintf(stderr,"bincache: [debug]: restart_article_leechers setting watch on %d\n",article->leechers[x]->fd);
        FD_SET(article->leechers[x]->fd,&g_watched_wfds);
      }
    } else {
      fprintf(stderr,"bincache: BUG: wanted to restart sending to a disconnected leecher.\n");
      article_detach_leecher(article,article->leechers[x]);
    }
  }
}

void article_handle_body_completion(struct desired_article *article)
{
  article->articlefd_wpos -= 2;
  lseek(article->articlefd,article->articlefd_wpos,SEEK_SET);
  ftruncate(article->articlefd,article->articlefd_wpos);
  //fprintf(stderr,"bincache: [debug]: article_handle_body_completion.\n");
  article->bodycomplete = 1;
  article->gettingbody = 0;
  restart_article_leechers(article);
  if (article->headerscomplete) {
    close(article->articlefd);
    char articlepartfn[PATH_MAX];
    snprintf(&articlepartfn[0],PATH_MAX,"%s/%s.articlepart",g_cache_dir,message_id_to_path_stem(article->message_id));
    articlepartfn[PATH_MAX-1] = '\0';
    char articlefn[PATH_MAX];
    snprintf(&articlefn[0],PATH_MAX,"%s/%s.article",g_cache_dir,message_id_to_path_stem(article->message_id));
    articlefn[PATH_MAX-1] = '\0';
    if (!rename(&articlepartfn[0],&articlefn[0])) {
      article->articlefd = open(&articlefn[0],O_BINARY|O_RDONLY);
      if (article->articlefd<0) {
        fprintf(stderr,"bincache: open() error %d (%s)\n",errno,strerror(errno));
      } else {
        article->articlelen = lseek(article->articlefd,0,SEEK_END);
	recache_desired_article_as_posessed(article);
      }
    } else {
      fprintf(stderr,"bincache: rename() error %d (%s)\n",errno,strerror(errno));
    }
  } else {
    close(article->articlefd);
    char bodypartfn[PATH_MAX];
    snprintf(&bodypartfn[0],PATH_MAX,"%s/%s.bodypart",g_cache_dir,message_id_to_path_stem(article->message_id));
    bodypartfn[PATH_MAX-1] = '\0';
    char bodyfn[PATH_MAX];
    snprintf(&bodyfn[0],PATH_MAX,"%s/%s.body",g_cache_dir,message_id_to_path_stem(article->message_id));
    bodyfn[PATH_MAX-1] = '\0';
    if (!rename(&bodypartfn[0],&bodyfn[0])) {
      article->articlefd = open(&bodyfn[0],O_BINARY|O_RDONLY);
      if (article->articlefd<0) {
        fprintf(stderr,"bincache: error %d (%s) opening %s\n",errno,
          strerror(errno), &bodyfn[0]);
      }
    } else {
      fprintf(stderr,"bincache: error %d (%s) renaming %s to %s\n",errno,
        strerror(errno),&bodypartfn[0],&bodyfn[0]);
    }
  }
}

void peer_handle_body_block(struct peer_connection *peer)
{
  /* If there's not room for it to have the .CRLF terminator, ignore. */
  if (peer->bufpos-peer->bufstart<3) return;
  int wrote;
  struct desired_article *article = lookup_desired_article(peer->message_id);
  if (!article) {
    fprintf(stderr,"bincache: article no longer desired.\n");
    return;
  }
  if (lseek(article->articlefd, article->articlefd_wpos, SEEK_SET)!=article->articlefd_wpos) {
    fprintf(stderr,"bincache: seek error %d (%s)\n", errno, strerror(errno));
    return;
  }
  /* We need to keep track of the place in the buffer where the current line
   * started, or it can misparse an overlong line.
   */
  while (peer->buflinelen>=0) {
    if (peer->block_at_crlf && peer->buf[peer->bufstart]=='.') {
      if (peer->buflinelen == 1) {
        article_handle_body_completion(article);
        peer->state = PEER_READY_FOR_COMMAND;
        peer->bufstart += 3;
        return;
      } else {
        peer->bufstart ++;
        peer->buflinelen --;
      }
    }
    wrote = write(article->articlefd, &peer->buf[peer->bufstart], peer->buflinelen+2);
    if (wrote!=peer->buflinelen+2) {
      fprintf(stderr,"bincache: write error %d (%s).\n",errno,strerror(errno));
      article_close_files(article);
      return;
    }
    article->articlefd_wpos += wrote;
    peer->bufstart += peer->buflinelen + 2;
    if (peer->bufstart>4096) {
      fprintf(stderr,"Sanity error. XXX\n");
      breakable();
    }
    peer->block_at_crlf = 1;
    peer_find_crlf(peer);
  }
  if (article->refcount<=0) {
    fprintf(stderr,"bincache: [BUG]: article reference count <= 0\n");
    breakable();
  }
  restart_article_leechers(article);
  if (article->refcount<0) {
    fprintf(stderr,"bincache: [BUG]: article reference count < 0\n");
    breakable();
  }
  if (peer->bufpos-peer->bufstart>5) {
    if (peer->block_at_crlf && peer->buf[peer->bufstart]=='.') {
      peer->bufstart++;
      peer->buflinelen --;
    }
    peer->block_at_crlf = 0;
    wrote = write(article->articlefd, &peer->buf[peer->bufstart], (peer->bufpos-peer->bufstart)-3);
    if (wrote!=(peer->bufpos-peer->bufstart)-3) {
      fprintf(stderr,"bincache: write error %d (%s)\n",errno,strerror(errno));
      return;
    }
    peer->bufstart += (peer->bufpos-peer->bufstart)-3;
    if (peer->bufstart>=BINCACHE_BUFFER_LEN) {
      fprintf(stderr,"XXX DECREMENTING!\n");
      assert(peer->bufstart<=BINCACHE_BUFFER_LEN);
      peer->bufpos = peer->bufstart = 0;
    }
    article->articlefd_wpos += wrote;
  }
}

void peer_handle_223(struct peer_connection *peer)
{
  int x = peer->bufstart+4;
  while (peer->buf[x]>='0' && peer->buf[x]<='9') {
    x++;
  }
  if (x<peer->bufpos && peer->buf[x]==' ') {
    x++;
    if (x<peer->bufpos && peer->buf[x]=='<') {
      x++;
      int message_id_start = x;
      while (x<peer->bufpos) {
        if (peer->buf[x]=='>') {
          peer->buf[x] = '\0';
          struct desired_article *article = lookup_desired_article(&peer->buf[message_id_start]);
	  if (article) {
            article_note_have_once(article,peer);
          } else {
	    fprintf(stderr,"bincache: no longer interested in article %s\n",&peer->buf[message_id_start]);
          }
        }
        x++;
      }
    }
  } else {
    // XXX note embedded protocol error.
  }
  if (peer->state==PEER_WAITING_FOR_STAT_RESPONSE) {
    peer->state = PEER_BIDI_READY_FOR_COMMAND;
  }
}

void peer_handle_stat_response(struct peer_connection *peer)
{
  peer_calculate_lag(peer);
  if (!memcmp(&peer->buf[peer->bufstart],"223 ",4)) {
    peer_handle_223(peer);
  } else if (!memcmp(&peer->buf[peer->bufstart],"430",3)) {
    struct desired_article *art = lookup_desired_article(peer->message_id);
    if (art) {
      article_note_donthave(art, peer);
    } else {
      fprintf(stderr,"bincache: debug: got a 430 response to a stat for %s, which we've forgotten we want.\n",peer->message_id);
    }
  }
  peer->state = PEER_READY_FOR_COMMAND;
}

int peer_discard_block(struct peer_connection *peer)
{
  if (!memcmp(&peer->buf[peer->bufstart],".\015\012",3)) {
    peer->state = PEER_READY_FOR_COMMAND;
    return 1;
  }
  return 0;
}

int null_rangle(char *buf, int len)
{
  int x = 0;
  while (x<len) {
    if (buf[x]=='>') {
      buf[x]='\0';
      return 1;
    }
    x++;
  }
  return 0;
}

void handle_bincache_221(struct peer_connection *peer)
{
  if (null_rangle(&peer->buf[peer->bufstart],peer->buflinelen)) {
    struct desired_article *article = lookup_desired_article(&peer->buf[peer->bufstart+7]);
    if (article) {
      article_note_have_once(article, peer);
      peer->state = PEER_RECEIVING_HEAD_BLOCK;
      peer->block_at_crlf = 1;
      peer_set_message_id(peer,&peer->buf[peer->bufstart+7]);
      //peer_attach_receiving_article(peer,article); /* TODO actually use this. */
    } else {
      fprintf(stderr,"bincache: Error: peer sent an article header we're no longer interested in.\n");
    }
  } else {
    fprintf(stderr,"bincache: Invalid 221 received.\n");
  }
}

void handle_bincache_222(struct peer_connection *peer)
{
  if (null_rangle(&peer->buf[peer->bufstart],peer->buflinelen)) {
    struct desired_article *article = lookup_desired_article(&peer->buf[peer->bufstart+7]);
    if (article) {
      article_note_have_once(article, peer);
      article->gettingbody = 1;
      peer->state = PEER_RECEIVING_BODY_BLOCK;
      peer->block_at_crlf = 1;
      peer_set_message_id(peer,&peer->buf[peer->bufstart+7]);
      //peer_attach_receiving_article(peer,article); /* TODO actually use this. */
    } else {
      fprintf(stderr,"bincache: Error: peer sent an article body we're no longer interested in.\n");
    }
  } else {
    fprintf(stderr,"bincache: Invalid 222 received.\n");
  }
}

void handle_bincache_430(struct peer_connection *peer)
{
  if (null_rangle(&peer->buf[peer->bufstart],peer->buflinelen)) {
    struct desired_article *article = lookup_desired_article(&peer->buf[peer->bufstart+5]);
    if (article) {
      article_note_donthave(article,peer);
      article_check_unavailability(article);
    }
    peer->state = PEER_BIDI_READY_FOR_COMMAND;
  }
}

void handle_bincache_225(struct peer_connection *peer)
{
  int x = peer->bufstart+4;
  int maxx = peer->buflinelen+peer->bufstart;
  while (x<maxx && peer->buf[x]!=' ') {
    x++;
  }
  if (x>=maxx) {
    fprintf(stderr,"bincache: malformed 225* received.\n");
    return;
  }
  peer->buf[x]='\0';
  x++;
  x++;
  if (null_rangle(&peer->buf[peer->bufstart],peer->buflinelen)) {
    article_t *article = lookup_desired_article(&peer->buf[x]);
    if (article) {
      article_note_have(article,peer);
      peer->message_id = strdup(&peer->buf[x]);
      peer_set_receiving_article(peer,article);
      if (!strcasecmp(&peer->buf[4],"newsgroups")) {
        peer->state = PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK;
      } else if (!strcasecmp(&peer->buf[4],"date")) {
        peer->state = PEER_RECEIVING_DATE_HDR_BLOCK;
      } else {
        fprintf(stderr,"bincache: Unknown header in 225*\n");
	peer->state = PEER_DISCARDING_BLOCK;
      }
      return;
    }
  }
  peer->state = PEER_DISCARDING_BLOCK;
}

void peer_handle_date_hdr_resp(struct peer_connection *peer)
{
  struct peer_connection *master = (peer->master ? peer->master : peer);
  if (!memcmp(&peer->buf[peer->bufstart],"225 ",4)) {
    peer->state = PEER_RECEIVING_DATE_HDR_BLOCK;
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"225*",4)) {
    handle_bincache_225(peer);
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"430 ",4)) {
    struct desired_article *article = lookup_desired_article(peer->message_id);
    /* Peer understands HDR, does not have the requested article. */
    if (article) {
      article_note_donthave(article,peer);
      article_check_unavailability(article);
    }
  } else if (!memcmp(&peer->buf[peer->bufstart],"430*<",5)) {
    handle_bincache_430(peer);
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"480 ",4)) {
    struct desired_article *article = lookup_desired_article(peer->message_id);
    if (article) article_remove_asked_note(article,peer);
    peer_authenticate(peer);
    return;
  } else if (!memcmp(&peer->buf[peer->bufstart],"500 ",4)) {
    master->hdr_date_not_supported = 1;
  } else if (!memcmp(&peer->buf[peer->bufstart],"503 ",4)) {
    master->hdr_date_not_supported = 1;
    struct desired_article *article = lookup_desired_article(peer->message_id);
    if (article) article_remove_asked_note(article,peer);
  } else {
    fprintf(stderr,"bincache: unexpected response to HDR command.\n");
    breakable();
  }
  if (proto_is_nntp(peer->prot)) {
    peer->state = PEER_READY_FOR_COMMAND;
  } else if (peer->state==PROTOCOL_BINCACHE_PEER) {
    peer->state = PEER_BIDI_READY_FOR_COMMAND;
  } else {
    fprintf(stderr,"bincache: Error: peer does not like HDR DATE.\n");
  }
  peer_writable_idle(peer);
}

void bincache_send_430(struct peer_connection *peer, const char *message_id)
{
  char buf[1024];
  int buflen = snprintf(&buf[0],1024,"430*<%s>\015\012",message_id);
  if (peer->trace) {
    fprintf(stderr,"bincache: [trace] %s [C] 430*<%s>\n",peer->name,message_id);
  }
  if (buflen>1024 || peer_conn_writebuf(peer,&buf[0],buflen)!=buflen) {
    fprintf(stderr,"bincache: Error writing buffer.\n");
    breakable();
    peer_connection_close(peer);
  }
}

void handle_bincache_stat(struct peer_connection *peer)
{
  char buf[1024];
  int buflen;
  char *message_id = &peer->buf[peer->bufstart+6];
  if (null_rangle(message_id,peer->buflinelen-6)) {
    char *articlefile = lookup_article(message_id);
    if (articlefile) {
      buflen = snprintf(&buf[0],1024,"223 0 <%s>\015\012",message_id);
      if (buflen<1024) {
	if (peer->trace) {
          fprintf(stderr,"bincache: [trace] %s [C] 223 0 <%s>\n",peer->name,message_id);
        }
        if (peer_conn_writebuf(peer,&buf[0],buflen)!=buflen) {
	  fprintf(stderr,"bincache: Error writing buffer.\n");
	  breakable();
	}
      }
    } else {
      bincache_send_430(peer,message_id);
    }
  } else {
    peer_conn_writes(peer,"501 syntax error.\015\012");
  }
}

void bincache_sendf_ss(struct peer_connection *peer, const char *fmt, const char *s1, const char *s2)
{
  char buf[1024];
  int buflen = snprintf(&buf[0],1024,fmt,s1,s2);
  if (buflen>1024 || peer_conn_writebuf(peer,&buf[0],buflen)!=buflen) {
    fprintf(stderr,"bincache: Error writing buffer.\n");
    breakable();
    peer_connection_close(peer);
  }
}

void bincache_sendf_s(struct peer_connection *peer, const char *fmt, const char *s)
{
  char buf[1024];
  int buflen = snprintf(&buf[0],1024,fmt,s);
  if (buflen>1024 || peer_conn_writebuf(peer,&buf[0],buflen)!=buflen) {
    fprintf(stderr,"bincache: Error writing buffer.\n");
    breakable();
    peer_connection_close(peer);
  }
}

void article_open(struct desired_article *article)
{
  if (article->articlefd>=0) {
    /* Assume that the article is already open if it's got an FD. */
    return;
  }
  char *fn = lookup_article(article->message_id);
  int fd = open(fn,O_RDONLY|O_BINARY);
  if (fd>=0) {
    article->articlefd = fd;
  } else {
    fprintf(stderr,"bincache: Error %d opening %s\n",errno,fn);
  }
  free(fn);
}

void handle_bincache_head(struct peer_connection *peer)
{
  if (null_rangle(&peer->buf[peer->bufstart],peer->buflinelen)) {
    char *message_id = &peer->buf[peer->bufstart+6];
    struct desired_article *article = lookup_posessed_article(&peer->buf[peer->bufstart+6]);
    if (article) {
      if (article->headerscomplete) {
        if (article->articlefd<0) {
	  article_open(article);
	}
	if (article->articlefd>=0) {
          article_init_from_cache(article);
          peer_attach_sending_article(peer,article);
          char buf[1024];
          snprintf(&buf[0],1023,"221 0 <%s>\015\012",message_id);
          buf[1023] = '\0';
          peer_conn_writes(peer,&buf[0]);
	  peer->write_block_atlf = 1;
          peer->state = PEER_SENDING_ARTICLE_HEAD;
          FD_SET(peer->fd,&g_watched_wfds);
	  peer->sending_article_rpos = 0;
          return;
        }
      }
    }
    char buf[1024];
    snprintf(&buf[0],1023,"430*<%s>\015\012",message_id);
    peer_conn_writes(peer,&buf[0]);
    return;
  } else {
    fprintf(stderr,"bincache: Error parsing bincache HEAD request.\n");
  }
}

void handle_bincache_body(struct peer_connection *peer)
{
  char *message_id = &peer->buf[peer->bufstart+6];
  if (null_rangle(message_id,peer->buflinelen-6)) {
    struct desired_article *article = lookup_posessed_article(message_id);
    if (!article) {
      char *articlefile = lookup_article(message_id);
      if (articlefile) {
        article = desired_article_alloc(message_id);
	article->articlefd = open(articlefile,O_RDONLY);
      }
    }
    if (article) {
      if (article->articlefd<0) {
	article_open(article);
      }
      if (article->articlefd>=0) {
        article_init_from_cache(article);
	peer_attach_sending_article(peer,article);
	bincache_sendf_s(peer,"222 0 <%s>\015\012",message_id);
	peer->state = PEER_SENDING_ARTICLE_BODY;
	peer->write_block_atlf = 1;
	peer->sending_article_rpos = article->articlefd_bodystart;
	FD_SET(peer->fd,&g_watched_wfds);
	return;
      }
    }
  }
  bincache_send_430(peer,message_id);
}

/* 
 * Returns the length of the message_id if it is valid.  Otherwise, returns 0.
 * TODO It's not sanity-checking quoting or escaping mechanisms.  Also, is
 * it between 3 and 250 _inclusive_ or _exclusive_?  I'm assuming inclusive.
 */
int validate_message_id(const char *message_id)
{
  int x = 0;
  while (message_id[x]) {
    if (message_id[x]<0x21 || message_id[x]>0x7e) return 0;
    x++;
  }
  if (x<3 || x>250) return 0;
  return x;
}

char *article_header(article_t *article, const char *hdr)
{
  if (article->headerscomplete) {
    int x = 0;
    int atlf = 1;
    int hdrlen = strlen(hdr);
    if (!strcasecmp(hdr,"newsgroups")) {
      char *ret = malloc(article->newsgroups_header_len+1); // set newsgroups header len
      memcpy(ret,article->newsgroups_header,article->newsgroups_header_len);
      ret[article->newsgroups_header_len] = '\0';
      return ret;
    }
    int maxx = article->headlen-(hdrlen+4);
    // XXX make sure this properly retreives the very last header.
    while (x<maxx) {
      if (atlf) {
        if (!strncasecmp(hdr,&article->head[x],hdrlen)) {
	  x+=hdrlen;
	  if (article->head[x]==':' && article->head[x+1]==' ') {
	    x += 2;
	    int hdrstart = x;
            while (x<maxx && article->head[x]!='\015') {
	      x++;
	    }
	    char *ret = malloc((x-hdrstart)+1);
            memcpy(ret,&article->head[hdrstart],x-hdrstart);
	    ret[x-hdrstart] = '\0';
	    return ret;
	  }
	  atlf = 0;
	}
      } else if (article->head[x]=='\012') {
        atlf=1;
      }
      x++;
    }
  }
  /* In theory, we could get headers from a partial article some of the time. 
   * in practice, we probably don't actually want to mess with that.
   */
  return NULL;
}

article_t *open_posessed_article(const char *message_id)
{
  article_t *ret = lookup_posessed_article(message_id);
  if (ret) return ret;
  char *filename = lookup_article(message_id);
  if (filename) {
    int fd = open(filename,O_RDWR);
    if (fd>=0) {
      ret = desired_article_alloc(message_id);
      g_num_posessed_articles++;
      g_posessed_articles = realloc(g_posessed_articles,sizeof(struct desired_article *)*g_num_posessed_articles);
      g_posessed_articles[g_num_posessed_articles-1] = ret;
      ret->articlefd = fd;
      article_init_from_cache(ret);
    }
  }
  return ret;
}

void peer_write_hdr_response(struct peer_connection *peer, const char *field, const char *message_id)
{
  struct desired_article *article = open_posessed_article(message_id);
  if (!article) {
    if (peer->prot == PROTOCOL_BINCACHE_PEER) {
      bincache_sendf_s(peer,"430*<%s>\015\012",message_id);
    } else {
      if (peer_conn_writebuf(peer,"430 \015\012",6)!=6) peer_close(peer);
      // XXX error-checking
    }
    return;
  }
  article_ref(article);
  int midlen = validate_message_id(message_id);
  if (midlen) {
    if (peer->prot == PROTOCOL_BINCACHE_PEER) {
      bincache_sendf_ss(peer,"225*%s <%s>\015\012",field,message_id);
      char *hdr = article_header(article,field);
      if (hdr) {
        bincache_sendf_s(peer,"0 %s\015\012.\015\012", hdr);
	free(hdr);
      } else {
        if (peer_conn_writebuf(peer,"0\015\012.\015\012",6)!=6) 
	  peer_close(peer);
	// XXX better error-checking.
      }
    } else {
      if (peer_conn_writebuf(peer,"500 \015\012",6)!=6) peer_close(peer);
      fprintf(stderr,"bincache: BUG: HDR not yet implemented for non-bincache peers.\n");
    }
  } else {
    if (peer_conn_writebuf(peer,"501 \015\012",6)!=6) peer_close(peer);
    // XXX TODO error-checking.
  }
  article_deref(article);
}

void handle_bincache_hdr(struct peer_connection *peer)
{
  if (peer->buf[peer->bufstart+3]==' ') {
    int x = peer->bufstart+4;
    while (peer->buf[x]!=' ' && x<peer->buflinelen) {
      x++;
    }
    if (x<peer->buflinelen+peer->bufstart) {
      peer->buf[x]='\0';
      char *field = &peer->buf[4];
      x++;
      if (peer->buf[x]=='<') {
        x++;
        char *message_id = &peer->buf[x];
        if (null_rangle(message_id,peer->buflinelen-x)) {
	  peer_write_hdr_response(peer,field,message_id);
	  return;
	}
      }
    }
    fprintf(stderr,"bincache: Malformed HDR request received.\n");
    peer_conn_writebuf(peer,"503 \015\012",6);
    return;
  } else if (peer->buf[peer->bufstart+3]!='\015') {
    peer_conn_writebuf(peer,"500 \015\012",6);
    fprintf(stderr,"bincache: unexpected command received.n");
    return;
  }
  peer_conn_writebuf(peer,"501 \015\012",6);
  fprintf(stderr,"bincache: HDR without args not allowed.\n");
}

void handle_bincache_wantrange(struct peer_connection *peer)
{
  /*
   * TODO: It might be nice to keep track of ranges desired and offer articles
   * in those to peers when they're complete.  This would imply an offering
   * protocol, which cannot be lifted from NNTP as we want the datestamp and
   * the message-ID.
   */
}

void handle_bincache_input(struct peer_connection *peer, char *orig_crlf)
{
  char *crlf = orig_crlf;
  peer_ref(peer);
  while (crlf) {
    switch (peer->state) {
      case PEER_DISCARDING_BLOCK:
        peer_discard_block(peer);
	break;
      case PEER_RECEIVING_HEAD_BLOCK:
      case PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT:
        peer_handle_head_block_line(peer);
        break;
      case PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK:
        peer_handle_hdr_block_line(peer);
	break;
      case PEER_RECEIVING_BODY_BLOCK:
        peer_handle_body_block(peer);
        peer_deref(peer);
        return;
        /* Return, since the body block handler parses multiple lines. */
        break;
      default:
	if (!memcmp(&peer->buf[peer->bufstart],"221 0 <",7)) {
          handle_bincache_221(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"222 0 <",7)) {
          handle_bincache_222(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"223 0 <",7)) {
          peer_handle_223(peer);
	} else if (!memcmp(&peer->buf[peer->bufstart],"225*",4)) {
	  handle_bincache_225(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"430*<",5)) {
          handle_bincache_430(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"631 ",4)) {
          /* Ignore the banners. */
	  if (peer->state==PEER_READING_BANNER) peer->state = PEER_READY_FOR_COMMAND;
        } else if (!memcmp(&peer->buf[peer->bufstart],"BODY <",6)) {
          handle_bincache_body(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"HEAD ",5)) {
          handle_bincache_head(peer);
	} else if (!memcmp(&peer->buf[peer->bufstart],"HDR",3)) {
	  handle_bincache_hdr(peer);
        } else if (!memcmp(&peer->buf[peer->bufstart],"STAT <",6)) {
          handle_bincache_stat(peer);
	} else if (!memcmp(&peer->buf[peer->bufstart],"WANTRANGE ",10)) {
	  handle_bincache_wantrange(peer);
        } else {
          if (peer->buf[peer->bufstart]<'0' || peer->buf[peer->bufstart]>'9') {
 	    fprintf(stderr,"Unexpected command received.\n");
 	    breakable();
            if (peer_conn_writebuf(peer,"500 \015\012",6)!=6) {
    	      peer_connection_close(peer);
	    }
  	  } else {
	    fprintf(stderr,"Unexpected numeric response received.\n");
  	    breakable();
          }
        }
    }
    peer->bufstart += peer->buflinelen+2;
    assert(peer->bufstart<=4096);
    crlf = peer_find_crlf(peer);
  }
  assert(peer->bufstart<=4096);
  peer_deref(peer);
}

void handle_peer_input(struct peer_connection *peer)
{
  peer_ref(peer);
  peer->lastread = time(NULL);
  if (peer->trace>1) fprintf(stderr,"bincache: [trace] set peer connection lastread to %d\n",(int)peer->lastread);
  int inputlen = peer->bufpos - peer->bufstart;
  char *crlf = find_crlf(&peer->buf[peer->bufstart],inputlen);
  if (crlf) {
    peer->buflinelen = crlf-&peer->buf[peer->bufstart];
    if (peer->trace) {
      if (peer->state!=PEER_RECEIVING_BODY_BLOCK || peer->trace>1) {
        *crlf = '\0';
        printf("bincache: [trace] %d:%s [S] %s\n",peer->num,peer->name,&peer->buf[peer->bufstart]);
        *crlf='\015';
      }
    }
  } else {
    peer->buflinelen = -1;
  }
  switch (peer->prot) {
    case PROTOCOL_BINCACHE_PEER:
      handle_bincache_input(peer,crlf);
      peer_deref(peer);
      return;
    default:
      // Ignore
      break;
  }
  while (crlf) {
    switch (peer->state) {
      case PEER_TCP_CONNECTING:
        peer->state = PEER_READING_BANNER;
        break;
      case PEER_READING_BANNER:
        peer_handle_banner(peer);
        break;
      case PEER_READY_FOR_COMMAND:
	breakable();
        peer_handle_ready_input(peer);
        break;
      case PEER_WAITING_FOR_STAT_RESPONSE:
        peer_handle_stat_response(peer);
        break;
      case PEER_WAITING_FOR_MODE_READER_RESP:
        peer_handle_mode_reader_resp(peer);
        break;
      case PEER_WAITING_FOR_AUTHINFO_USER_RESP:
        peer_handle_authinfo_user_resp(peer);
        break;
      case PEER_WAITING_FOR_AUTHINFO_PASS_RESP:
        peer_handle_authinfo_pass_resp(peer);
        break;
      case PEER_WAITING_FOR_NEWSGROUPS_HDR:
        peer_handle_newsgroups_hdr(peer);
        break;
      case PEER_RECEIVING_NEWSGROUPS_HDR_BLOCK:
        peer_handle_hdr_block_line(peer);
        break;
      case PEER_WAITING_FOR_HEAD_RESP:
        peer_handle_head_resp(peer);
        break;
      case PEER_RECEIVING_HEAD_BLOCK:
      case PEER_RECEIVING_HEAD_BLOCK_NEWSGROUPS_CONT:
        peer_handle_head_block_line(peer);
        break;
      case PEER_DISCARDING_BLOCK:
        if (peer_discard_block(peer)) peer_writable_idle(peer);
        break;
      case PEER_WAITING_FOR_BODY_RESP:
        peer_handle_body_resp(peer);
        break;
      case PEER_RECEIVING_BODY_BLOCK:
        peer_handle_body_block(peer);
        /* Return, since the body block handler parses multiple lines. */
        peer_deref(peer);
        return; 
        break;
      case PEER_WAITING_FOR_DATE_HDR:
        peer_handle_date_hdr_resp(peer);
	peer_deref(peer);
        return;
      default:
        printf("bincache: %s: input in invalid peer state %d\n", peer->name, peer->state);
    }
    if (crlf-&peer->buf[0]==peer->bufpos) {
      peer->bufstart = peer->bufpos = 0;
    } else {
      peer->bufstart = (crlf-&peer->buf[0])+2;
    }
    crlf = peer_find_crlf(peer);
  }
  switch (peer->state) {
    case PEER_TCP_CONNECTING:
      peer->state = PEER_READING_BANNER;
      break;
    case PEER_RECEIVING_BODY_BLOCK:
      peer_handle_body_block(peer);
      break;
    default:
      printf("Debug: 4397\n");
      // XXX uterminated
      break;
  }
  peer_deref(peer);
}

void handle_peer_readable(struct peer_connection *peer)
{
  if (peer->state==PEER_TLS_HANDSHAKING) {
    fprintf(stderr,"bincache: [debug]: handshaking...\n");
    int tlserr = gnutls_handshake(peer->tls_session);
    if (tlserr>=0) {
      if (peer->prot == PROTOCOL_BINCACHE_PEER) {
        peer_conn_writes(peer,"631 \015\012");
      }
      peer->state = PEER_READING_BANNER;
      return;
    } else if ((tlserr==GNUTLS_E_INTERRUPTED) || (tlserr==GNUTLS_E_AGAIN)) {
      if (gnutls_record_get_direction(peer->tls_session)) {
        FD_CLR(peer->fd,&g_watched_rfds);
        FD_SET(peer->fd,&g_watched_wfds);
        fprintf(stderr,"bincache: [debug]: handshake in progress, select reset to write.\n");
      } else {
        FD_CLR(peer->fd,&g_watched_wfds);
        FD_SET(peer->fd,&g_watched_rfds);
      }
      return;
    } else {
      fprintf(stderr,"bincache: handle_peer_readable: gnutls_handshake err:\n\t");
      gnutls_perror(tlserr);
      peer_close(peer);
      return;
    }
  }
  if (peer->bufpos == BINCACHE_BUFFER_LEN) {
    if (peer->bufstart<BINCACHE_BUFFER_LEN) {
      memmove(&peer->buf[0],&peer->buf[peer->bufstart],BINCACHE_BUFFER_LEN-peer->bufstart);
      peer->bufpos -= peer->bufstart;
    } else {
      peer->bufpos = 0;
    }
    peer->bufstart = 0;
    // XXX buffer-length blocks will still choke it.
  }
  int nread;
  if (peer->bufpos==peer->bufstart) {
    peer->bufpos = peer->bufstart = 0;
  }
  if (!peer->tls) {
    nread = recv(peer->fd,&peer->buf[peer->bufpos],BINCACHE_BUFFER_LEN-peer->bufpos,0);
  } else {
    nread = gnutls_record_recv(peer->tls_session,&peer->buf[peer->bufpos],BINCACHE_BUFFER_LEN-peer->bufpos);
    if (nread<=5) breakable(); // XXX debug;
  }
  if (nread>0) {
    peer->bufpos += nread;
    if (peer->bufpos==4096) {
      if (peer->bufstart<4096) {
        memmove(&peer->buf[0],&peer->buf[peer->bufstart],peer->bufpos-peer->bufstart);
	peer->bufpos -= peer->bufstart;
	peer->bufstart = 0;
      }
    }
    handle_peer_input(peer);
  } else if (nread<0) {
    if (nread!=GNUTLS_E_INTERRUPTED && nread!=GNUTLS_E_AGAIN) {
      printf("bincache: recv() error %d, %s, %d, %s.\n",errno,strerror(errno), sockerrno, strerror(sockerrno));
      // TODO note read errors?
      peer_connection_close(peer);
    } else {
      fprintf(stderr,"bincache: [debug]: nonblocking read must resume.\n");
      return;
    }
  } else {
    printf("bincache: connection closed by peer %s.\n",peer->name);
    // TODO note closure
    peer_connection_close(peer);
  }
}

int handle_client_sending_writebuf(struct client_connection *conn)
{
  int towrite = conn->writebuflen - conn->writebufpos;
  if (towrite>4096) {
    towrite = 4096;
  } else if (towrite<0) {
    fprintf(stderr,"bincache: error: want to write a negative number.\n");
  }
  int wrote = send(conn->fd,&conn->writebuf[conn->writebufpos],towrite,0);
  if (wrote<0) {
    fprintf(stderr,"bincache: send failed %d (%s).\n", errno, strerror(errno));
    client_close(conn);
  } else {
    conn->writebufpos += wrote;
    if (conn->writebufpos>=conn->writebuflen) {
      if (conn->writebufpos>conn->writebuflen) {
        fprintf(stderr,"bincache: BUG: wrote %d bytes off end of writebuf.\n", conn->writebufpos);
      }
      return 1;
    }
  }
  return 0;
}

int client_send_dotstuffed(struct client_connection *conn, char *buf, int len)
{
  char stuffed[4096];
  int x = 0;
  int y = 0;
  assert(len<=4096);
  int atlf = conn->write_block_atlf;
  while (x<len) {
    stuffed[y++] = buf[x];
    if (buf[x]!='\012') {
      if (buf[x]=='.' && atlf) {
        stuffed[y++] = '.';
      }
      atlf = 0;
    } else {
      atlf = 1;
    }
    x++;
  }
  conn->write_block_atlf = atlf;
  if (find_crcr(&stuffed[0],y)) {
    fprintf(stderr,"bincache: BUG: CRCR found.\n");
  }
  if (send(conn->fd,&stuffed[0],y,0)==y) return len;
  return -1;
}

int client_send_file_offset(struct client_connection *client, int offset)
{
  int towrite;
  char buf[2048];
  int real_rpos = client->articlefd_rpos + offset;
  if (client->article->gettingbody) {
    if (client->article->articlefd_wpos>real_rpos) {
      towrite = client->article->articlefd_wpos - real_rpos;
      if (towrite>2048) towrite = 2048;
      if (lseek(client->article->articlefd,real_rpos,SEEK_SET)==real_rpos) {
        if (read(client->article->articlefd,&buf[0],towrite)==towrite) {
          int sent = client_send_dotstuffed(client, &buf[0], towrite);
          if (sent>0) {
            client->articlefd_rpos += sent;
          } else if (sent==0) {
            fprintf(stderr,"bincache: error: sent 0 bytes?\n");
          } else {
            fprintf(stderr,"bincache: send error %d (%s)\n",errno,strerror(errno));
          }
        } else {
          fprintf(stderr,"bincache: read error\n");
	  fprintf(stderr,"[%d,%s]\n",errno,strerror(errno));
	  breakable();
	  assert(0);
        }
      } else {
        fprintf(stderr,"bincache: lseek error %d (%s) sending article body\n",errno,strerror(errno));
      }
    } else {
//      fprintf(stderr,"bincache: cleared wfd watch on %d\n",client->fd);
      FD_CLR(client->fd,&g_watched_wfds);
    }
  } else if (client->article->bodycomplete) {
    towrite = client->article->articlelen - real_rpos;
    if (towrite<=0) {
      if (!(towrite==0 || towrite==-2)) fprintf(stderr,"Unexpected towrite value.\n");
      return 1;
    }
    if (towrite>2048) towrite = 2048;
    if (lseek(client->article->articlefd, real_rpos, SEEK_SET)==real_rpos) {
      if (read(client->article->articlefd, &buf[0], towrite)==towrite) {
        int sent = client_send_dotstuffed(client, &buf[0], towrite);
        if (sent>0) {
          client->articlefd_rpos += sent;
        } else {
          fprintf(stderr,"bincache: send error %d (%s)\n",errno,strerror(errno));
        }
      } else {
        fprintf(stderr,"bincache: read error %d (%s)\n", errno, strerror(errno));
      }
    } else {
      fprintf(stderr,"bincache: lseek error %d (%s)\n",errno,strerror(errno));
    }
  } else {
    //fprintf(stderr,"bincache: [error] clearing watched wfd %d\n",client->fd);
    //Not really an error, there's simply no progress.
    FD_CLR(client->fd,&g_watched_wfds);
  }
  return 0;
}

int handle_client_sending_article_body(struct client_connection *client)
{
  return client_send_file_offset(client, client->article->articlefd_bodystart);
}

int handle_client_sending_article(struct client_connection *client)
{
  return client_send_file_offset(client, 0);
}

void handle_client_writable(struct client_connection *conn)
{
  switch (conn->state) {
    case CLIENT_SENDING_ARTICLE_HEAD:
      if (handle_client_sending_writebuf(conn)) {
        conn->state = CLIENT_SENDING_ARTICLE_BODY;
        client_conn_writes(conn,"\015\012");
        conn->write_block_atlf = 1;
        conn->articlefd_rpos = 0;
      }
      break;
    case CLIENT_SENDING_ARTICLE_BODY:
      if (handle_client_sending_article_body(conn)) {
        //fprintf(stderr,"bincache: [debug]: peer connection sent article body.\n");
        conn->state = CLIENT_WAITING_FOR_COMMAND;
        FD_CLR(conn->fd,&g_watched_wfds);
        client_conn_writes(conn,"\015\012.\015\012");
        article_detach_leecher(conn->article, conn);
      }
      break;
    case CLIENT_SENDING_ARTICLE:
      if (handle_client_sending_article(conn)) {
        //fprintf(stderr,"bincache: [debug]: peer connection sent article entire.\n");
        conn->state = CLIENT_WAITING_FOR_COMMAND;
        FD_CLR(conn->fd,&g_watched_wfds);
        client_conn_writes(conn,"\015\012.\015\012");
        article_detach_leecher(conn->article, conn);
      }
      break;
    default:
      fprintf(stderr,"bincache: BUG: writability noticed in unknown state.  Clearing writability trap.\n");
      FD_CLR(conn->fd,&g_watched_wfds);
      break;
  }
}

void peer_detach_sending_article(struct peer_connection *peer)
{
  article_deref(peer->sending_article);
  peer->sending_article_rpos = 0;
  peer->sending_article = NULL;
}

void peer_send_article_head(struct peer_connection *peer)
{
  struct desired_article *article = peer->sending_article;
  char buf[4096];
  if (lseek(article->articlefd,peer->sending_article_rpos,SEEK_SET)!=peer->sending_article_rpos) {
    fprintf(stderr,"bincache: lseek error.\n");
    breakable();
    peer_connection_close(peer);
    return;
  }
  int numtoread = (peer->sending_article->articlefd_bodystart-4)-peer->sending_article_rpos;
  if (numtoread>4096) {
    numtoread = 4096;
  }
  int nread = read(article->articlefd,&buf[0],numtoread); /* XXX there was something being debugged around here... */
  if (nread>0) {
    char stuffed[8192];
    int ipos = 0;
    int opos = 0;
    int write_block_atlf = peer->write_block_atlf;
    while (ipos<nread) {
      if (write_block_atlf) {
        if (buf[ipos]=='.') {
          stuffed[opos++]='.';
	}
	write_block_atlf = 0;
      }
      if (buf[ipos]=='\012') write_block_atlf = 1;
      stuffed[opos++] = buf[ipos++];
    }
    if (peer_conn_writebuf(peer, &stuffed[0], opos)!=opos) {
      fprintf(stderr,"bincache: incomplete write.\n");
      breakable();
      peer_connection_close(peer);
    }
    peer->write_block_atlf = write_block_atlf;
    peer->sending_article_rpos += nread;
  } else if (nread==0) {
    fprintf(stderr,"bincache: [debug] Done sending article to peer.\n");
    if (peer_conn_writebuf(peer,"\015\012.\015\012",5)!=5) {
      fprintf(stderr,"bincache: incomplete write.\n");
      breakable();
      peer_connection_close(peer);
      return;
    }
    peer->write_block_atlf = 0;
    if (peer->prot==PROTOCOL_BINCACHE_PEER) {
      peer->state = PEER_BIDI_READY_FOR_COMMAND;
    } else {
      fprintf(stderr,"ERROR: sending header to a non-bincache peer?\n");
    }
    peer_detach_sending_article(peer);
  } else {
    fprintf(stderr,"bincache: read() error.\n");
    peer_detach_sending_article(peer);
    breakable();
    return;
  }
}

void peer_send_article_body(struct peer_connection *peer)
{
  struct desired_article *article = peer->sending_article;
  char buf[4096];
  if (lseek(article->articlefd,peer->sending_article_rpos,SEEK_SET)!=peer->sending_article_rpos) {
    fprintf(stderr,"bincache: lseek error.\n");
    breakable();
    peer_connection_close(peer);
    return;
  }
  int nread = read(article->articlefd,&buf[0],409); // XXX was 4096, testing to see if it's a blocksize-related error.
  if (nread>0) {
    char stuffed[8192];
    int ipos = 0;
    int opos = 0;
    int write_block_atlf = peer->write_block_atlf;
    while (ipos<nread) {
      if (write_block_atlf) {
        if (buf[ipos]=='.') {
          stuffed[opos++]='.';
	}
	write_block_atlf = 0;
      }
      if (buf[ipos]=='\012') write_block_atlf = 1;
      stuffed[opos++] = buf[ipos++];
    }
    if (peer_conn_writebuf(peer, &stuffed[0], opos)!=opos) {
      fprintf(stderr,"bincache: incomplete write.\n");
      breakable();
      peer_connection_close(peer);
    }
    peer->write_block_atlf = write_block_atlf;
    peer->sending_article_rpos += nread;
  } else if (nread==0) {
    fprintf(stderr,"bincache: [debug] Done sending article to peer.\n");
    if (peer_conn_writebuf(peer,"\015\012.\015\012",5)!=5) {
      fprintf(stderr,"bincache: incomplete write.\n");
      breakable();
      peer_connection_close(peer);
      return;
    }
    peer->write_block_atlf = 0;
    peer->state = PEER_READY_FOR_COMMAND;
    peer_detach_sending_article(peer);
  } else {
    fprintf(stderr,"bincache: read() error.\n");
    peer_detach_sending_article(peer);
    breakable();
    return;
  }
}

void handle_peer_writable(struct peer_connection *peer)
{
  if (peer->state==PEER_READY_FOR_COMMAND) {
    if (!peer_writable_idle(peer)) {
      //fprintf(stderr,"bincache: [debug]: handle_peer_writable clearing watched wfds.\n");
      FD_CLR(peer->fd,&g_watched_wfds);
    }
  } else if (peer->state == PEER_TCP_CONNECTING) {
    int err;
    socklen_t errsize = sizeof(err);
    if (!getsockopt(peer->fd,SOL_SOCKET,SO_ERROR,&err,&errsize)) {
      FD_CLR(peer->fd,&g_watched_wfds);
      if (peer->tls) {
        peer_connect_tls_2(peer);
      } else {
        peer->state = PEER_READING_BANNER;
      }
    } else {
      fprintf(stderr,"bincache: [error]: getsockopt() failed.\n");
      peer_close(peer);
    }
  } else if (peer->state == PEER_SENDING_ARTICLE_BODY) {
    peer_send_article_body(peer);
  } else if (peer->state == PEER_SENDING_ARTICLE_HEAD) {
    peer_send_article_head(peer);
  } else {
    fprintf(stderr,"bincache: [debug]: handle_peer_writable clearing watched wfds(2).\n");
    breakable();
    FD_CLR(peer->fd,&g_watched_wfds);
  }
}

void clean_stale_desired_articles(void)
{
  int x = g_num_desired_articles;
  while (x) {
    x--;
    if (g_desired_articles[x]->refcount==0 && g_desired_articles[x]->last_activity+300<time(NULL)) {
      forget_article(g_desired_articles[x]);
    }
  }
}

void check_selected_client_fds(fd_set *rfds, fd_set *wfds)
{
  int x = g_num_client_connections;
  while (x) {
    x--;
    if (FD_ISSET(g_client_connections[x]->fd,rfds)) {
      handle_client_readable(g_client_connections[x]);
    } else if (FD_ISSET(g_client_connections[x]->fd,wfds)) {
      handle_client_writable(g_client_connections[x]);
    } else if (g_client_connections[x]->state == CLIENT_WAITING_FOR_COMMAND && g_client_connections[x]->bufstart<g_client_connections[x]->bufpos) {
      handle_client_readable(g_client_connections[x]);
    }
    if (g_client_connections[x]->fd==-1) {
      client_deref(g_client_connections[x]);
      g_client_connections[x] = g_client_connections[--g_num_client_connections];
      // Not sure, but I think that this could enable someone to disrupt
      // bincache by establishing lots and lots of connections and closing
      // them right away.  On the other hand, it's not really intended to
      // be set up open to the general public.
    }
  }
}

void check_selected_peer_fds(fd_set *rfds, fd_set *wfds)
{
  int x = g_num_peer_connections;
  while (x) {
    x--;
    if (g_peer_connections[x]->fd>=0) {
      peer_ref(g_peer_connections[x]);
      if (FD_ISSET(g_peer_connections[x]->fd,rfds)) {
        handle_peer_readable(g_peer_connections[x]);
        if (g_peer_connections[x]->state==PEER_READY_FOR_COMMAND || g_peer_connections[x]->state==PEER_BIDI_READY_FOR_COMMAND) {
          peer_writable_idle(g_peer_connections[x]);
        } else if (time(NULL)-g_peer_connections[x]->lastread>1300) {
          /* XXX long timeout. */
          fprintf(stderr,"bincache: [debug] %d-%d>30\n",(int)time(NULL), (int)g_peer_connections[x]->lastread);
          peer_remove(g_peer_connections[x]);
        }
      } else if (FD_ISSET(g_peer_connections[x]->fd,wfds)) {
        handle_peer_writable(g_peer_connections[x]);
      }
      peer_deref(g_peer_connections[x]);
    }
  }
}

void check_selected_cacheplex_fds(fd_set *rfds, fd_set *wfds)
{
  int x = g_num_cacheplex_connections;
  while (x) {
    x--;
    if (g_cacheplex_connections[x]->fd>=0) {
      if (FD_ISSET(g_cacheplex_connections[x]->fd,rfds)) {
        handle_peer_readable(g_cacheplex_connections[x]);
        if (g_cacheplex_connections[x]->state==PEER_READY_FOR_COMMAND) {
          FD_SET(g_cacheplex_connections[x]->fd,&g_watched_wfds);
        }
      } else if (FD_ISSET(g_cacheplex_connections[x]->fd,wfds)) {
        handle_peer_writable(g_cacheplex_connections[x]);
      }
    }
    // XXX check for timeouting Cacheplex connections.
  }
}

int ready_for_command(struct peer_connection *peer)
{
  if (peer->state == PEER_READY_FOR_COMMAND) return 1;
  if (peer->state == PEER_BIDI_READY_FOR_COMMAND) return 1;
  return 0;
}

int main(int argc, char **argv)
{
  init_bincache();
  while (1) {
    fd_set rfds = g_watched_rfds;
    fd_set wfds = g_watched_wfds;
    struct timeval tv;
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    int readable = select(g_high_fd, &rfds, &wfds, NULL, &tv);
    if (readable>0) {
      check_selected_client_fds(&rfds,&wfds);
      check_selected_peer_fds(&rfds,&wfds);
      check_selected_cacheplex_fds(&rfds,&wfds);
      if (FD_ISSET(g_nntp_listen_socket,&rfds)) {
        handle_nntp_connection(g_nntp_listen_socket);
      }
      if (g_bincache_listen_socket>=0 && FD_ISSET(g_bincache_listen_socket,&rfds)) {
        handle_bincache_peer_connection(g_bincache_listen_socket);
      }
    } else if (readable<0) {
      fprintf(stderr,"bincache: select error %d (%s).\n",errno,strerror(errno));
      #ifndef WIN32
      sleep(1);
      #endif
    } else {
      clean_stale_desired_articles();
    }
    /* Trying moving the peer_writable_idle out of the select-success loop. */
    int x = g_num_peer_connections;
    while (x) {
      x--;
      if (ready_for_command(g_peer_connections[x])) {
        peer_writable_idle(g_peer_connections[x]);
      }
    }
    /* Client input is checked regardless of readable/writable state */
    x = g_num_client_connections;
    while (x) {
      x--;
      if (g_client_connections[x]->state == CLIENT_WAITING_FOR_COMMAND && g_client_connections[x]->bufstart<g_client_connections[x]->bufpos) {
        handle_client_input(g_client_connections[x]);
      }
    }
  }
#ifdef WIN32
  WSACleanup();
#endif
}
